import logging

from PySide import QtGui

import spreads
from spreads.plugin import HookPlugin
import gui
import gui_rc

logger = logging.getLogger('spreadsplug.gui')


class GuiCommand(HookPlugin):
    @classmethod
    def add_command_parser(cls, rootparser):
        guiparser = rootparser.add_parser(
            'gui', help="Start the GUI wizard")
        guiparser.set_defaults(func=wizard)


def wizard(args):
    logger.debug("Starting GUI")
    app = QtGui.QApplication([])
    wizard = gui.SpreadsWizard(spreads.config)
    wizard.show()
    app.exec_()



import logging
import operator
import os

import wand.api
import wand.image
from concurrent import futures

from spreads.plugin import HookPlugin

logger = logging.getLogger('spreadsplug.colorcorrect')


def correct_colors(img_path, factors):
    logger.debug("Correcting color of \"{0}\"".format(img_path))
    for channel, factor in zip(('red', 'green', 'blue'), factors):
        logger.debug("Correcting {0} channel by {1}"
                     .format(channel, factor))
        with wand.image.Image(filename=img_path) as img:
            wand.api.libmagick.MagickEvaluateImageChannel(
                img.wand, wand.image.CHANNELS[channel],
                wand.image.EVALUATE_OPS.index('multiply'),
                factor)
            img.save(filename=img_path)


class ColorCorrectionPlugin(HookPlugin):
    def process(self, path):
        path = os.path.join(path, 'raw')
        logger.debug("Starting color correction...")
        # Get the gray card's RGB values from configuration
        true_colors = (float(self.config['colorcorrect']['true_red']
                             .get(int)),
                       float(self.config['colorcorrect']['true_green']
                             .get(int)),
                       float(self.config['colorcorrect']['true_blue']
                             .get(int))
                       )
        # We assume that the first two images shot were the gray card
        images = sorted([os.path.join(path, x) for x in os.listdir(path)])
        factors_left = map(operator.div, true_colors,
                           self._get_color(images[0]))
        factors_right = map(operator.div, true_colors,
                            self._get_color(images[1]))

        with futures.ProcessPoolExecutor() as executor:
            # Don't correct the pictures with the gray cards
            for idx, img in enumerate(images[3:]):
                if not idx % 2:
                    factors = factors_left
                else:
                    factors = factors_right
                executor.submit(correct_colors, img, factors)

    def _get_color(self, img_path):
        """ Get average color of image's core 500x500 area. """
        with wand.image.Image(filename=img_path) as img:
            img.crop(left=img.size[0]/2-250,
                     top=img.size[1]/2-250,
                     width=500, height=500)           # crop rectangle
            img.resize(1, 1)
            color = (img[0][0].red_int8, img[0][0].green_int8,
                     img[0][0].blue_int8)
        return color



# -*- coding: utf-8 -*-

from __future__ import division, unicode_literals

import logging
import os
import subprocess

from spreads.plugin import HookPlugin
from spreads.util import MissingDependencyException, find_in_path

if not find_in_path('pdfbeads'):
    raise MissingDependencyException("Could not find executable `pdfbeads` in"
                                     " $PATH. Please install the appropriate"
                                     " package(s)!")

logger = logging.getLogger('spreadsplug.pdfbeads')


class PDFBeadsPlugin(HookPlugin):
    def output(self, path):
        logger.info("Assembling PDF.")
        img_dir = os.path.join(path, 'done')
        pdf_file = os.path.join('..', 'out',
                                "{0}.pdf".format(os.path.basename(path)))
        img_files = [x
                     for x in sorted(os.listdir(img_dir))
                     if os.path.splitext(x)[1].lower() == '.tif']
        cmd = ["pdfbeads", "-d"] + img_files + ["-o", pdf_file]
        logger.debug("Running " + " ".join(cmd))
        # NOTE: pdfbeads only finds *html files for the text layer in the
        #       working directory...
        os.chdir(img_dir)
        _ = subprocess.check_output(cmd, stderr=subprocess.STDOUT)



# -*- coding: utf-8 -*-

# Copyright (c) 2013 Johannes Baiter. All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
spreads CLI code.
"""

from __future__ import division, unicode_literals, print_function

import argparse
import logging
import os
import sys
import time

import colorama
import spreads.confit as confit

import spreads.workflow as workflow
from spreads import config
from spreads.plugin import get_devices, get_pluginmanager
from spreads.util import DeviceException, ColourStreamHandler

# Kudos to http://stackoverflow.com/a/1394994/487903
try:
    from msvcrt import getch
except ImportError:
    def getch():
        """ Wait for keypress on stdin.

        :returns: unicode -- Value of character that was pressed

        """
        import tty
        import termios
        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            return sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old)


def configure(args=None):
    for orientation in ('left', 'right'):
        print("Please connect and turn on the device labeled \'{0}\'"
              .format(orientation))
        print(colorama.Fore.BLUE + "Press any key when ready.")
        _ = getch()
        devs = get_devices()
        if len(devs) > 1:
            raise DeviceException("Please ensure that only one device is"
                                  " turned on!")
        if not devs:
            raise DeviceException("No device found!")
        devs[0].set_orientation(orientation)
        print(colorama.Fore.GREEN + "Configured \'{0}\' device.".format(orientation))
        print("Please turn off the device.")
        print(colorama.Fore.BLUE + "Press any key when ready.")
        getch()


def capture(args=None, devices=[]):
    if not devices:
        print("Starting capture workflow, please connect and turn on the"
              " devices.")
        print(colorama.Fore.BLUE + "Press any key to continue.")
        getch()
        print("Detecting devices.")
        devices = get_devices()
        if len(devices) != 2:
            raise DeviceException("Please connect and turn on two"
                                  " pre-configured devices! ({0} were"
                                  " found)".format(len(devices)))
        print(colorama.Fore.GREEN + "Found {0} devices!".format(len(devices)))
        if any(not x.orientation for x in devices):
            raise DeviceException("At least one of the devices has not been"
                                  " properly configured, please re-run the"
                                  " program with the \'configure\' option!")
    # Set up for capturing
    print("Setting up devices for capturing.")
    workflow.prepare_capture(devices)
    # Start capture loop
    print(colorama.Fore.BLUE + "Press 'b' to capture.")
    shot_count = 0
    start_time = time.time()
    pages_per_hour = 0
    capture_keys = config['capture']['capture_keys'].as_str_seq()
    while True:
        if not getch().lower() in capture_keys:
            break
        workflow.capture(devices)
        shot_count += len(devices)
        pages_per_hour = (3600/(time.time() - start_time))*shot_count
        status = ("\rShot {0} pages [{1:.0f}/h]"
                  .format(colorama.Fore.GREEN + unicode(shot_count), pages_per_hour))
        sys.stdout.write(status)
        sys.stdout.flush()
    workflow.finish_capture(devices)
    sys.stdout.write("\rShot {0} pages in {1:.1f} minutes, average speed was"
                     " {2:.0f} pages per hour"
                     .format(colorama.Fore.GREEN + str(shot_count),
                             (time.time() - start_time)/60, pages_per_hour))
    sys.stdout.flush()


def download(args=None, path=None):
    if args and args.path:
        path = args.path
    devices = get_devices()
    status_str = "Downloading {0} images from devices"
    if config['download']['keep'].get(bool) or config['keep'].get(bool):
        status_str = status_str.format("and deleting ")
    else:
        status_str = status_str.format("")
    print(colorama.Fore.GREEN + status_str)
    workflow.download(devices, path)


def postprocess(args=None, path=None):
    if args and args.path:
        path = args.path
    workflow.process(path)


def output(args=None, path=None):
    if args and args.path:
        path = args.path
    workflow.output(path)


def wizard(args):
    # TODO: Think about how we can make this more dynamic, i.e. get list of
    #       options for plugin with a description for each entry
    path = args.path
    print("Please connect and turn on the devices.")
    print(colorama.Fore.BLUE + "Press any key to continue.")
    getch()
    print(colorama.Fore.GREEN + "Detecting devices.")
    devices = get_devices()
    if any(not x.orientation for x in devices):
        print(colorama.Fore.YELLOW + "Devices not yet configured!")
        print(colorama.Fore.BLUE + "Please turn both devices off."
                          " Press any key when ready.")
        while True:
            try:
                configure()
                break
            except DeviceException as e:
                print(e)

    print(colorama.Fore.GREEN +
          "==========================\n",
          "Starting capturing process\n",
          "==========================")
    capture(devices=devices)

    print(colorama.Fore.GREEN +
          "=========================\n",
          "Starting download process\n"
          "=========================")
    download(path=path)

    print(colorama.Fore.GREEN +
          "=======================\n"
          "Starting postprocessing\n"
          "=======================")
    postprocess(path=path)

    print(colorama.Fore.GREEN +
          "=================\n",
          "Generating output\n"
          "=================")
    output(path=path)


def setup_parser():
    pluginmanager = get_pluginmanager()
    parser = argparse.ArgumentParser(
        description="Scanning Tool for  DIY Book Scanner")
    subparsers = parser.add_subparsers()

    parser.add_argument(
        '--verbose', '-v', dest="verbose", action="store_true")

    config_parser = subparsers.add_parser(
        'configure', help="Perform initial configuration of the devices.")
    config_parser.set_defaults(func=configure)

    capture_parser = subparsers.add_parser(
        'capture', help="Start the capturing workflow")
    capture_parser.set_defaults(func=capture)
    # Add arguments from plugins
    pluginmanager.map(lambda x, y, z: x.plugin.add_arguments(y, z),
                      'capture', capture_parser)

    download_parser = subparsers.add_parser(
        'download', help="Download scanned images.")
    download_parser.add_argument(
        "path", help="Path where scanned images are to be stored")
    download_parser.add_argument(
        "--keep", "-k", dest="keep", action="store_true",
        help="Keep files on devices after download")
    download_parser.set_defaults(func=download)
    # Add arguments from plugins
    pluginmanager.map(lambda x, y, z: x.plugin.add_arguments(y, z),
                      'download', download_parser)

    postprocess_parser = subparsers.add_parser(
        'postprocess',
        help="Postprocess scanned images.")
    postprocess_parser.add_argument(
        "path", help="Path where scanned images are stored")
    postprocess_parser.add_argument(
        "--jobs", "-j", dest="jobs", type=int, default=None,
        metavar="<int>", help="Number of concurrent processes")
    postprocess_parser.set_defaults(func=postprocess)
    # Add arguments from plugins
    pluginmanager.map(lambda x, y, z: x.plugin.add_arguments(y, z),
                      'postprocess', postprocess_parser)

    output_parser = subparsers.add_parser(
        'output',
        help="Generate output files.")
    output_parser.add_argument(
        "path", help="Path where scanned and postprocessed images are stored")
    output_parser.set_defaults(func=output)
    # Add arguments from plugins
    pluginmanager.map(lambda x, y, z: x.plugin.add_arguments(y, z),
                      'output', output_parser)

    wizard_parser = subparsers.add_parser(
        'wizard', help="Interactive mode")
    wizard_parser.add_argument(
        "path", help="Path where scanned images are to be stored")
    wizard_parser.set_defaults(func=wizard)

    pluginmanager.map(lambda x, y: x.plugin.add_command_parser(y),
                      subparsers)
    return parser


def main():
    # Initialize color support
    colorama.init()
    # Set to ERROR so we can see errors during plugin loading.
    logging.basicConfig(loglevel=logging.ERROR)
    config.read()
    cfg_path = os.path.join(config.config_dir(), confit.CONFIG_FILENAME)
    if not os.path.exists(cfg_path):
        config.dump(filename=cfg_path)

    parser = setup_parser()
    args = parser.parse_args()
    config.set_args(args)

    loglevel = config['loglevel'].as_choice({
        'none':     logging.NOTSET,
        'info':     logging.INFO,
        'debug':    logging.DEBUG,
        'warning':  logging.WARNING,
        'error':    logging.ERROR,
        'critical': logging.CRITICAL,
    })
    if args.verbose:
        loglevel = logging.DEBUG

    # Set up logger
    logger = logging.getLogger()
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    handler = ColourStreamHandler()
    handler.setLevel(loglevel)
    handler.setFormatter(logging.Formatter("%(name)s: %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(loglevel)

    args.func(args)

    # Deinitialize color support
    colorama.deinit()



# This file is part of Confit.
# Copyright 2013, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Worry-free YAML configuration files.
"""
from __future__ import unicode_literals
import platform
import os
import pkgutil
import re
import sys
import yaml
import types
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

UNIX_DIR_VAR = 'XDG_CONFIG_HOME'
UNIX_DIR_FALLBACK = '~/.config'
WINDOWS_DIR_VAR = 'APPDATA'
WINDOWS_DIR_FALLBACK = '~\\AppData\\Roaming'
MAC_DIR = '~/Library/Application Support'

CONFIG_FILENAME = 'config.yaml'
DEFAULT_FILENAME = 'config_default.yaml'
ROOT_NAME = 'root'

YAML_TAB_PROBLEM = "found character '\\t' that cannot start any token"


# Utilities.

PY3 = sys.version_info[0] == 3
STRING = str if PY3 else unicode
BASESTRING = str if PY3 else basestring
NUMERIC_TYPES = (int, float) if PY3 else (int, float, long)
TYPE_TYPES = (type,) if PY3 else (type, types.ClassType)

def iter_first(sequence):
    """Get the first element from an iterable or raise a ValueError if
    the iterator generates no values.
    """
    it = iter(sequence)
    try:
        if PY3:
            return next(it)
        else:
            return it.next()
    except StopIteration:
        raise ValueError()


# Exceptions.

class ConfigError(Exception):
    """Base class for exceptions raised when querying a configuration.
    """

class NotFoundError(ConfigError):
    """A requested value could not be found in the configuration trees.
    """

class ConfigTypeError(ConfigError, TypeError):
    """The value in the configuration did not match the expected type.
    """

class ConfigValueError(ConfigError, ValueError):
    """The value in the configuration is illegal."""

class ConfigReadError(ConfigError):
    """A configuration file could not be read."""
    def __init__(self, filename, reason=None):
        self.filename = filename
        self.reason = reason

        message = 'file {0} could not be read'.format(filename)
        if isinstance(reason, yaml.scanner.ScannerError) and \
                reason.problem == YAML_TAB_PROBLEM:
            # Special-case error message for tab indentation in YAML markup.
            message += ': found tab character at line {0}, column {1}'.format(
                reason.problem_mark.line + 1,
                reason.problem_mark.column + 1,
            )
        elif reason:
            # Generic error message uses exception's message.
            message += ': {0}'.format(reason)

        super(ConfigReadError, self).__init__(message)


# Views and sources.

class ConfigSource(OrderedDict):
    """A dictionary augmented with metadata about the source of the
    configuration.
    """
    def __init__(self, value, filename=None, default=False):
        super(ConfigSource, self).__init__(value)
        if filename is not None and not isinstance(filename, BASESTRING):
            raise TypeError('filename must be a string or None')
        self.filename = filename
        self.default = default

    def __repr__(self):
        return 'ConfigSource({0}, {1}, {2})'.format(
            super(ConfigSource, self).__repr__(),
            repr(self.filename),
            repr(self.default)
        )

    @classmethod
    def of(self, value):
        """Given either a dictionary or a `ConfigSource` object, return
        a `ConfigSource` object. This lets a function accept either type
        of object as an argument.
        """
        if isinstance(value, ConfigSource):
            return value
        elif isinstance(value, dict):
            return ConfigSource(value)
        else:
            raise TypeError('source value must be a dict')

class ConfigView(object):
    """A configuration "view" is a query into a program's configuration
    data. A view represents a hypothetical location in the configuration
    tree; to extract the data from the location, a client typically
    calls the ``view.get()`` method. The client can access children in
    the tree (subviews) by subscripting the parent view (i.e.,
    ``view[key]``).
    """

    name = None
    """The name of the view, depicting the path taken through the
    configuration in Python-like syntax (e.g., ``foo['bar'][42]``).
    """

    def resolve(self):
        """The core (internal) data retrieval method. Generates (value,
        source) pairs for each source that contains a value for this
        view. May raise ConfigTypeError if a type error occurs while
        traversing a source.
        """
        raise NotImplementedError

    def first(self):
        """Returns a (value, source) pair for the first object found for
        this view. This amounts to the first element returned by
        `resolve`. If no values are available, a NotFoundError is
        raised.
        """
        pairs = self.resolve()
        try:
            return iter_first(pairs)
        except ValueError:
            raise NotFoundError("{0} not found".format(self.name))

    def add(self, value):
        """Set the *default* value for this configuration view. The
        specified value is added as the lowest-priority configuration
        data source.
        """
        raise NotImplementedError

    def set(self, value):
        """*Override* the value for this configuration view. The
        specified value is added as the highest-priority configuration
        data source.
        """
        raise NotImplementedError

    def root(self):
        """The RootView object from which this view is descended.
        """
        raise NotImplementedError

    def __repr__(self):
        return '<ConfigView: %s>' % self.name

    def __getitem__(self, key):
        """Get a subview of this view."""
        return Subview(self, key)

    def __setitem__(self, key, value):
        """Create an overlay source to assign a given key under this
        view.
        """
        self.set({key: value})

    def set_args(self, namespace):
        """Overlay parsed command-line arguments, generated by a library
        like argparse or optparse, onto this view's value.
        """
        args = {}
        for key, value in namespace.__dict__.items():
            if value is not None:  # Avoid unset options.
                args[key] = value
        self.set(args)

    # Magical conversions. These special methods make it possible to use
    # View objects somewhat transparently in certain circumstances. For
    # example, rather than using ``view.get(bool)``, it's possible to
    # just say ``bool(view)`` or use ``view`` in a conditional.

    def __str__(self):
        """Gets the value for this view as a byte string."""
        return str(self.get())

    def __unicode__(self):
        """Gets the value for this view as a unicode string. (Python 2
        only.)
        """
        return unicode(self.get())

    def __nonzero__(self):
        """Gets the value for this view as a boolean. (Python 2 only.)
        """
        return self.__bool__()

    def __bool__(self):
        """Gets the value for this view as a boolean. (Python 3 only.)
        """
        return bool(self.get())

    # Dictionary emulation methods.

    def keys(self):
        """Returns a list containing all the keys available as subviews
        of the current views. This enumerates all the keys in *all*
        dictionaries matching the current view, in contrast to
        ``view.get(dict).keys()``, which gets all the keys for the
        *first* dict matching the view. If the object for this view in
        any source is not a dict, then a ConfigTypeError is raised. The
        keys are ordered according to how they appear in each source.
        """
        keys = []

        for dic, _ in self.resolve():
            try:
                cur_keys = dic.keys()
            except AttributeError:
                raise ConfigTypeError(
                    '{0} must be a dict, not {1}'.format(
                        self.name, type(dic).__name__
                    )
                )

            for key in cur_keys:
                if key not in keys:
                    keys.append(key)

        return keys

    def items(self):
        """Iterates over (key, subview) pairs contained in dictionaries
        from *all* sources at this view. If the object for this view in
        any source is not a dict, then a ConfigTypeError is raised.
        """
        for key in self.keys():
            yield key, self[key]

    def values(self):
        """Iterates over all the subviews contained in dictionaries from
        *all* sources at this view. If the object for this view in any
        source is not a dict, then a ConfigTypeError is raised.
        """
        for key in self.keys():
            yield self[key]

    # List/sequence emulation.

    def all_contents(self):
        """Iterates over all subviews from collections at this view from
        *all* sources. If the object for this view in any source is not
        iterable, then a ConfigTypeError is raised. This method is
        intended to be used when the view indicates a list; this method
        will concatenate the contents of the list from all sources.
        """
        for collection, _ in self.resolve():
            try:
                it = iter(collection)
            except TypeError:
                raise ConfigTypeError(
                    '{0} must be an iterable, not {1}'.format(
                        self.name, type(collection).__name__
                    )
                )
            for value in it:
                yield value

    # Validation and conversion.

    def get(self, typ=None):
        """Returns the canonical value for the view, checked against the
        passed-in type. If the value is not an instance of the given
        type, a ConfigTypeError is raised. May also raise a
        NotFoundError.
        """
        value, _ = self.first()

        if typ is not None:
            if not isinstance(typ, TYPE_TYPES):
                raise TypeError('argument to get() must be a type')

            if not isinstance(value, typ):
                raise ConfigTypeError(
                    "{0} must be of type {1}, not {2}".format(
                        self.name, typ.__name__, type(value).__name__
                    )
                )

        return value

    def as_filename(self):
        """Get a string as a normalized filename, made absolute and with
        tilde expanded. If the value comes from a default source, the
        path is considered relative to the application's config
        directory. If it comes from another file source, the filename is
        expanded as if it were relative to that directory. Otherwise, it
        is relative to the current working directory.
        """
        path, source = self.first()
        if not isinstance(path, BASESTRING):
            raise ConfigTypeError('{0} must be a filename, not {1}'.format(
                self.name, type(path).__name__
            ))
        path = os.path.expanduser(STRING(path))

        if source.default:
            # From defaults: relative to the app's directory.
            path = os.path.join(self.root().config_dir(), path)

        elif source.filename is not None:
            # Relative to source filename's directory.
            path = os.path.join(os.path.dirname(source.filename), path)

        return os.path.abspath(path)

    def as_choice(self, choices):
        """Ensure that the value is among a collection of choices and
        return it. If `choices` is a dictionary, then return the
        corresponding value rather than the value itself (the key).
        """
        value = self.get()

        if value not in choices:
            raise ConfigValueError(
                '{0} must be one of {1}, not {2}'.format(
                    self.name, repr(list(choices)), repr(value)
                )
            )

        if isinstance(choices, dict):
            return choices[value]
        else:
            return value

    def as_number(self):
        """Ensure that a value is of numeric type."""
        value = self.get()
        if isinstance(value, NUMERIC_TYPES):
            return value
        raise ConfigTypeError(
            '{0} must be numeric, not {1}'.format(
                self.name, type(value).__name__
            )
        )

    def as_str_seq(self):
        """Get the value as a list of strings. The underlying configured
        value can be a sequence or a single string. In the latter case,
        the string is treated as a white-space separated list of words.
        """
        value = self.get()
        if isinstance(value, bytes):
            value = value.decode('utf8', 'ignore')

        if isinstance(value, STRING):
            return value.split()
        else:
            try:
                return list(value)
            except TypeError:
                raise ConfigTypeError(
                    '{0} must be a whitespace-separated string or '
                    'a list'.format(self.name)
                )

    def flatten(self):
        """Create a hierarchy of OrderedDicts containing the data from
        this view, recursively reifying all views to get their
        represented values.
        """
        od = OrderedDict()
        for key, view in self.items():
            try:
                od[key] = view.flatten()
            except ConfigTypeError:
                od[key] = view.get()
        return od

class RootView(ConfigView):
    """The base of a view hierarchy. This view keeps track of the
    sources that may be accessed by subviews.
    """
    def __init__(self, sources):
        """Create a configuration hierarchy for a list of sources. At
        least one source must be provided. The first source in the list
        has the highest priority.
        """
        self.sources = list(sources)
        self.name = ROOT_NAME

    def add(self, obj):
        self.sources.append(ConfigSource.of(obj))

    def set(self, value):
        self.sources.insert(0, ConfigSource.of(value))

    def resolve(self):
        return ((OrderedDict(s), s) for s in self.sources)

    def clear(self):
        """Remove all sources from this configuration."""
        del self.sources[:]

    def root(self):
        return self

class Subview(ConfigView):
    """A subview accessed via a subscript of a parent view."""
    def __init__(self, parent, key):
        """Make a subview of a parent view for a given subscript key.
        """
        self.parent = parent
        self.key = key

        # Choose a human-readable name for this view.
        if isinstance(self.parent, RootView):
            self.name = ''
        else:
            self.name = self.parent.name
            if not isinstance(self.key, int):
                self.name += '.'
        if isinstance(self.key, int):
            self.name += '#{0}'.format(self.key)
        elif isinstance(self.key, BASESTRING):
            self.name += '{0}'.format(self.key)
        else:
            self.name += '{0}'.format(repr(self.key))

    def resolve(self):
        for collection, source in self.parent.resolve():
            try:
                value = collection[self.key]
            except IndexError:
                # List index out of bounds.
                continue
            except KeyError:
                # Dict key does not exist.
                continue
            except TypeError:
                # Not subscriptable.
                raise ConfigTypeError(
                    "{0} must be a collection, not {1}".format(
                        self.parent.name, type(collection).__name__
                    )
                )
            yield value, source

    def set(self, value):
        self.parent.set({self.key: value})

    def add(self, value):
        self.parent.add({self.key: value})

    def root(self):
        return self.parent.root()


# Config file paths, including platform-specific paths and in-package
# defaults.

# Based on get_root_path from Flask by Armin Ronacher.
def _package_path(name):
    """Returns the path to the package containing the named module or
    None if the path could not be identified (e.g., if
    ``name == "__main__"``).
    """
    loader = pkgutil.get_loader(name)
    if loader is None or name == '__main__':
        return None

    if hasattr(loader, 'get_filename'):
        filepath = loader.get_filename(name)
    else:
        # Fall back to importing the specified module.
        __import__(name)
        filepath = sys.modules[name].__file__

    return os.path.dirname(os.path.abspath(filepath))

def config_dirs():
    """Returns a list of user configuration directories to be searched.
    """
    if platform.system() == 'Darwin':
        paths = [UNIX_DIR_FALLBACK, MAC_DIR]
    elif platform.system() == 'Windows':
        if WINDOWS_DIR_VAR in os.environ:
            paths = [os.environ[WINDOWS_DIR_VAR]]
        else:
            paths = [WINDOWS_DIR_FALLBACK]
    else:
        # Assume Unix.
        paths = [UNIX_DIR_FALLBACK]
        if UNIX_DIR_VAR in os.environ:
            paths.insert(0, os.environ[UNIX_DIR_VAR])

    # Expand and deduplicate paths.
    out = []
    for path in paths:
        path = os.path.abspath(os.path.expanduser(path))
        if path not in out:
            out.append(path)
    return out


# YAML loading.

class Loader(yaml.SafeLoader):
    """A customized YAML loader. This loader deviates from the official
    YAML spec in a few convenient ways:

    - All strings as are Unicode objects.
    - All maps are OrderedDicts.
    - Strings can begin with % without quotation.
    """
    # All strings should be Unicode objects, regardless of contents.
    def _construct_unicode(self, node):
        return self.construct_scalar(node)

    # Use ordered dictionaries for every YAML map.
    # From https://gist.github.com/844388
    def construct_yaml_map(self, node):
        data = OrderedDict()
        yield data
        value = self.construct_mapping(node)
        data.update(value)

    def construct_mapping(self, node, deep=False):
        if isinstance(node, yaml.MappingNode):
            self.flatten_mapping(node)
        else:
            raise yaml.constructor.ConstructorError(
                None, None,
                'expected a mapping node, but found %s' % node.id,
                node.start_mark
            )

        mapping = OrderedDict()
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                raise yaml.constructor.ConstructorError(
                    'while constructing a mapping',
                    node.start_mark, 'found unacceptable key (%s)' % exc,
                    key_node.start_mark
                )
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping

    # Allow bare strings to begin with %. Directives are still detected.
    def check_plain(self):
        plain = super(Loader, self).check_plain()
        return plain or self.peek() == '%'

Loader.add_constructor('tag:yaml.org,2002:str', Loader._construct_unicode)
Loader.add_constructor('tag:yaml.org,2002:map', Loader.construct_yaml_map)
Loader.add_constructor('tag:yaml.org,2002:omap', Loader.construct_yaml_map)

def load_yaml(filename):
    """Read a YAML document from a file. If the file cannot be read or
    parsed, a ConfigReadError is raised.
    """
    try:
        with open(filename, 'r') as f:
            return yaml.load(f, Loader=Loader)
    except (IOError, yaml.error.YAMLError) as exc:
        raise ConfigReadError(filename, exc)


# YAML dumping.

class Dumper(yaml.SafeDumper):
    """A PyYAML Dumper that represents OrderedDicts as ordinary mappings
    (in order, of course).
    """
    # From http://pyyaml.org/attachment/ticket/161/use_ordered_dict.py
    def represent_mapping(self, tag, mapping, flow_style=None):
        value = []
        node = yaml.MappingNode(tag, value, flow_style=flow_style)
        if self.alias_key is not None:
            self.represented_objects[self.alias_key] = node
        best_style = False
        if hasattr(mapping, 'items'):
            mapping = list(mapping.items())
        for item_key, item_value in mapping:
            node_key = self.represent_data(item_key)
            node_value = self.represent_data(item_value)
            if not (isinstance(node_key, yaml.ScalarNode)
                    and not node_key.style):
                best_style = False
            if not (isinstance(node_value, yaml.ScalarNode)
                    and not node_value.style):
                best_style = False
            value.append((node_key, node_value))
        if flow_style is None:
            if self.default_flow_style is not None:
                node.flow_style = self.default_flow_style
            else:
                node.flow_style = best_style
        return node

    def represent_list(self, data):
        """If a list has less than 4 items, represent it in inline style
        (i.e. comma separated, within square brackets).
        """
        node = super(Dumper, self).represent_list(data)
        length = len(data)
        if self.default_flow_style is None and length < 4:
            node.flow_style = True
        elif self.default_flow_style is None:
            node.flow_style = False
        return node

    def represent_bool(self, data):
        """Represent bool as 'yes' or 'no' instead of 'true' or 'false'.
        """
        if data:
            value = 'yes'
        else:
            value = 'no'
        return self.represent_scalar('tag:yaml.org,2002:bool', value)

    def represent_none(self, data):
        """Represent a None value with nothing instead of 'none'.
        """
        return self.represent_scalar('tag:yaml.org,2002:null', '')

Dumper.add_representer(OrderedDict, Dumper.represent_dict)
Dumper.add_representer(bool, Dumper.represent_bool)
Dumper.add_representer(type(None), Dumper.represent_none)
Dumper.add_representer(list, Dumper.represent_list)

def restore_yaml_comments(data, default_data):
    """Scan default_data for comments (we include empty lines in our
    definition of comments) and place them before the same keys in data.
    Only works with comments that are on one or more own lines, i.e.
    not next to a yaml mapping.
    """
    comment_map = dict()
    default_lines = iter(default_data.splitlines())
    for line in default_lines:
        if not line:
            comment = "\n"
        elif line.startswith("#"):
            comment = "{0}\n".format(line)
        else:
            continue
        while True:
            line = next(default_lines)
            if line and not line.startswith("#"):
                break
            comment += "{0}\n".format(line)
        key = line.split(':')[0].strip()
        comment_map[key] = comment
    out_lines = iter(data.splitlines())
    out_data = ""
    for line in out_lines:
        key = line.split(':')[0].strip()
        if key in comment_map:
            out_data += comment_map[key]
        out_data += "{0}\n".format(line)
    return out_data


# Main interface.

class Configuration(RootView):
    def __init__(self, appname, modname=None, read=True):
        """Create a configuration object by reading the
        automatically-discovered config files for the application for a
        given name. If `modname` is specified, it should be the import
        name of a module whose package will be searched for a default
        config file. (Otherwise, no defaults are used.) Pass `False` for
        `read` to disable automatic reading of all discovered
        configuration files. Use this when creating a configuration
        object at module load time and then call the `read` method
        later.
        """
        super(Configuration, self).__init__([])
        self.appname = appname
        self.modname = modname

        self._env_var = '{0}DIR'.format(self.appname.upper())

        if read:
            self.read()

    def _search_dirs(self):
        """Yield directories that will be searched for configuration
        files for this application.
        """
        # Application's environment variable.
        if self._env_var in os.environ:
            path = os.environ[self._env_var]
            yield os.path.abspath(os.path.expanduser(path))

        # Standard configuration directories.
        for confdir in config_dirs():
            yield os.path.join(confdir, self.appname)

    def _user_sources(self):
        """Generate `ConfigSource` objects for each user configuration
        file in the program's search directories.
        """
        for appdir in self._search_dirs():
            filename = os.path.join(appdir, CONFIG_FILENAME)
            if os.path.isfile(filename):
                yield ConfigSource(load_yaml(filename) or {}, filename)

    def _default_source(self):
        """Return the default-value source for this program or `None` if
        it does not exist.
        """
        if self.modname:
            pkg_path = _package_path(self.modname)
            if pkg_path:
                filename = os.path.join(pkg_path, DEFAULT_FILENAME)
                if os.path.isfile(filename):
                    return ConfigSource(load_yaml(filename), filename, True)

    def read(self, user=True, defaults=True):
        """Find and read the files for this configuration and set them
        as the sources for this configuration. To disable either
        discovered user configuration files or the in-package defaults,
        set `user` or `defaults` to `False`.
        """
        if user:
            for source in self._user_sources():
                self.add(source)
        if defaults:
            source = self._default_source()
            if source:
                self.add(source)

    def config_dir(self):
        """Get the path to the directory containing the highest-priority
        user configuration. If no user configuration is present, create a
        suitable directory before returning it.
        """
        dirs = list(self._search_dirs())

        # First, look for an existent configuration file.
        for appdir in dirs:
            if os.path.isfile(os.path.join(appdir, CONFIG_FILENAME)):
                return appdir

        # As a fallback, create the first-listed directory name.
        appdir = dirs[0]
        if not os.path.isdir(appdir):
            os.makedirs(appdir)
        return appdir

    def dump(self, filename=None, full=True):
        """Dump the Configuration object to a YAML file.

        The order of the keys is determined from the default
        configuration file. All keys not in the default configuration
        will be appended to the end of the file.

        :param filename:  The file to dump the configuration to, or None
                          if the YAML string should be returned instead
        :type filename:   unicode
        :param full:      Dump settings that don't differ from the defaults
                          as well
        """
        out_dict = OrderedDict()
        default_conf = next(x for x in self.sources if x.default)
        try:
            default_keys = list(default_conf.keys())
        except AttributeError:
            default_keys = []
        new_keys = [x for x in self.keys() if not x in default_keys]
        out_keys = default_keys + new_keys
        for key in out_keys:
            # Skip entries unchanged from default config
            if (not full and key in default_keys
                    and self[key].get() == default_conf[key]):
                continue
            try:
                out_dict[key] = self[key].flatten()
            except ConfigTypeError:
                out_dict[key] = self[key].get()

        yaml_out = yaml.dump(out_dict, Dumper=Dumper,
                             default_flow_style=None, indent=4,
                             width=1000)

        # Restore comments to the YAML text.
        with open(default_conf.filename, 'r') as fp:
            default_data = fp.read()
        yaml_out = restore_yaml_comments(yaml_out, default_data)

        # Return the YAML or write it to a file.
        if filename is None:
            return yaml_out
        else:
            with open(filename, 'w') as fp:
                fp.write(yaml_out)


class LazyConfig(Configuration):
    """A Configuration at reads files on demand when it is first
    accessed. This is appropriate for using as a global config object at
    the module level.
    """
    def __init__(self, appname, modname=None):
        super(LazyConfig, self).__init__(appname, modname, False)
        self._materialized = False  # Have we read the files yet?
        self._lazy_prefix = []  # Pre-materialization calls to set().
        self._lazy_suffix = []  # Calls to add().

    def read(self, user=True, defaults=True):
        self._materialized = True
        super(LazyConfig, self).read(user, defaults)

    def resolve(self):
        if not self._materialized:
            # Read files and unspool buffers.
            self.read()
            self.sources += self._lazy_suffix
            self.sources[:0] = self._lazy_prefix
        return super(LazyConfig, self).resolve()

    def add(self, value):
        super(LazyConfig, self).add(value)
        if not self._materialized:
            # Buffer additions to end.
            self._lazy_suffix += self.sources
            del self.sources[:]

    def set(self, value):
        super(LazyConfig, self).set(value)
        if not self._materialized:
            # Buffer additions to beginning.
            self._lazy_prefix[:0] = self.sources
            del self.sources[:]



from spreads.plugin import BaseCamera


class MockCamera(BaseCamera):
    @classmethod
    def match(cls, vendor_id, product_id):
        pass

    def __init__(self, bus=0, device=0, orientation='left'):
        self._port = (bus, device)
        self.orientation = orientation

    def set_orientation(self, orientation):
        self.orientation = orientation

    def delete_files(self):
        pass

    def download_files(self, path):
        pass

    def set_record_mode(self):
        pass

    def get_zoom(self):
        return self._zoom or 0

    def set_zoom(self, level):
        self._zoom = level

    def disable_flash(self):
        pass

    def set_iso(self, iso_value):
        pass

    def disable_ndfilter(self):
        pass

    def shoot(self, shutter_speed, iso_value):
        pass

    def play_sound(self, sound_num):
        pass



# -*- coding: utf-8 -*-

from __future__ import division

import logging
import os

import wand.image
from concurrent import futures
from pexif import JpegFile

from spreads.plugin import HookPlugin

logger = logging.getLogger('spreadsplug.autorotate')


def rotate_image(path, rotation):
    with wand.image.Image(filename=path) as img:
        logger.debug("Rotating image \'{0}\' by {1} degrees"
                     .format(path, rotation))
        if img.height > img.width:
            logger.info("Image already rotated, skipping...")
        else:
            img.rotate(rotation)
            img.save(filename=path)
    # Update EXIF orientation
    img = JpegFile.fromFile(path)
    if not img.exif.primary.Orientation == [1]:
        logger.debug("Updating EXIF information for image '{0}'".format(path))
        img.exif.primary.Orientation = [1]
        img.writeFile(path)


class AutoRotatePlugin(HookPlugin):
    @classmethod
    def add_arguments(cls, command, parser):
        if command == 'postprocess':
            parser.add_argument("--rotate-inverse", "-ri",
                                dest="rotate_inverse", action="store_true",
                                default=False,
                                help="Rotate left pages CCW, right pages CW"
                                " (use when first page comes from right"
                                " camera)")

    def process(self, path):
        img_dir = os.path.join(path, 'raw')
        # Silence Wand logger
        (logging.getLogger("wand")
                .setLevel(logging.ERROR))
        logger.info("Rotating images in {0}".format(img_dir))
        with futures.ProcessPoolExecutor() as executor:
            for imgpath in os.listdir(img_dir):
                try:
                    img = JpegFile.fromFile(os.path.join(img_dir, imgpath))
                    if img.exif.primary.Orientation == [8]:
                        rotation = (self.config['autorotate']
                                    ['left'].get(int))
                    elif img.exif.primary.Orientation == [6]:
                        rotation = (self.config['autorotate']
                                    ['right'].get(int))
                    elif img.exif.primary.Orientation == [1]:
                        # Already rotated, so we skip it
                        continue
                    else:
                        raise Exception("Invalid value for orientation: {0}"
                                        .format(img.exif.primary.Orientation))
                except Exception as exc:
                    logger.warn("Cannot determine rotation for image {0}!"
                                .format(imgpath))
                    logger.exception(exc)
                    continue
                if self.config['rotate_inverse'].get(bool):
                    rotation *= -1
                logger.debug("Orientation for \"{0}\" is {1}"
                             .format(imgpath, rotation))
                executor.submit(
                    rotate_image,
                    os.path.join(img_dir, imgpath),
                    rotation=rotation
                )



# -*- coding: utf-8 -*-

# Copyright (c) 2013 Johannes Baiter. All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
spreads workflow steps.
"""

from __future__ import division, unicode_literals

import logging
import os
import sys
import time

from concurrent.futures import ThreadPoolExecutor

from spreads import config
from spreads.plugin import get_pluginmanager

logger = logging.getLogger('spreads.workflow')


def prepare_capture(devices):
    logger.debug("Preparing capture.")
    with ThreadPoolExecutor(len(devices)) as executor:
        futures = []
        logger.debug("Preparing capture in devices")
        for dev in devices:
            futures.append(executor.submit(dev.prepare_capture))
    if any(x.exception() for x in futures):
        exc = next(x for x in futures if x.exception()).exception()
        logger.error("There was an exception while preparing for capture",
                     exc_info=sys.exc_info(exc))
    for ext in get_pluginmanager():
        logger.debug("Running prepare_capture hooks")
        ext.obj.prepare_capture(devices)


def capture(devices):
    logger.info("Triggering capture.")
    with ThreadPoolExecutor(len(devices)) as executor:
        futures = []
        logger.debug("Sending capture command to devices")
        for dev in devices:
            futures.append(executor.submit(dev.capture))
    if any(x.exception() for x in futures):
        exc = next(x for x in futures if x.exception()).exception()
        raise exc
    logger.debug("Running capture hooks")
    for ext in get_pluginmanager():
        ext.obj.capture(devices)


def finish_capture(devices):
    logger.debug("Running finish_capture hooks")
    for ext in get_pluginmanager():
        ext.obj.finish_capture(devices)


def download(devices, path):
    keep = config['download']['keep'].get(bool) or config['keep'].get(bool)
    logger.info("Downloading images from devices to {0}, files will "
                "{1} remain on the devices."
                .format(path, ("not" if not keep else "")))
    if not os.path.exists(path):
        logger.debug("Creating project directory")
        os.mkdir(path)
    out_paths = [os.path.join(path, x.orientation) for x in devices]
    if out_paths:
        logger.debug("Creating camera directories")

    # Flag for when the images are already present
    skip_download = False
    for subpath in out_paths:
        if not os.path.exists(subpath):
            os.mkdir(subpath)
        else:
            logger.info("Images already present, skipping download")
            skip_download = True
    if not skip_download:
        with ThreadPoolExecutor(len(devices)) as executor:
            logger.debug("Starting download process")
            futures = []
            for dev in devices:
                futures.append(executor.submit(
                    dev.download_files,
                    os.path.join(path, dev.orientation))
                )
            if any(x.exception() for x in futures):
                exc = next(x for x in futures if x.exception()).exception()
                raise exc
    logger.debug("Running download hooks")
    for ext in get_pluginmanager():
        ext.obj.download(devices, path)
    # NOTE: Just to be safe...
    time.sleep(5)
    if not keep:
        logger.info("Deleting images from devices")
        with ThreadPoolExecutor(len(devices)) as executor:
            logger.debug("Starting delete process")
            for dev in devices:
                executor.submit(dev.delete_files)
        logger.debug("Running delete hooks")
        for ext in get_pluginmanager():
            ext.obj.delete(devices)


def process(path):
    logger.info("Starting postprocessing...")
    logger.debug("Running process hooks")
    for ext in get_pluginmanager():
        ext.obj.process(path)
    logger.info("Done with postprocessing!")


def output(path):
    logger.info("Generating output files...")
    logger.debug("Running output hooks")
    out_path = os.path.join(path, 'out')
    if not os.path.exists(out_path):
        os.mkdir(out_path)
    for ext in get_pluginmanager():
        ext.obj.output(path)
    logger.info("Done generating output files!")



# -*- coding: utf-8 -*-
import logging
import os
import time
from fractions import Fraction
from math import log

import pexif
import pyptpchdk
from PIL import Image
from pyptpchdk import PTPError

from spreads.plugin import DevicePlugin
from spreads.util import DeviceException


class PTPDevice(object):
    """ Small wrapper class around pyptpchdk for easier access to PTP
        functions.

    """
    STATUS_RUN = 1
    STATUS_MSG = 2

    def __init__(self, usbdevice):
        self._orientation = None
        self.logger = logging.getLogger('PTPDevice')
        self._device = pyptpchdk.PTPDevice(usbdevice.bus, usbdevice.address)

    def __del__(self):
        del(self._device)

    def execute_lua(self, script, wait=True, get_result=False, timeout=256):
        """ Executes a Lua script on the camera.

        :param script: The Lua script
        :type script: str
        :param wait: Wait for the script to complete
        :type wait: bool
        :param get_result: Return the script's return value
        :type get_result: bool
        :return: The script's return value (tuple) if get_result was True,
                 or None

        """
        # Wrap the script in return if the script doesn't return by itself
        if get_result and not "return" in script:
            script = "return({0})".format(script)
        retries = 0
        while True:
            if retries >= 3:
                raise DeviceException("An error occured while executing"
                                      " \"{0}\"".format(script))
            self.logger.debug("Executing script: \"{0}\"".format(script))
            try:
                script_id = self._device.chdkExecLua(script)
            except Exception:
                self.logger.warn("Script raised an error.")
                self.logger.debug("Details:", exc_info=True)
                self.logger.warn("Retrying in 5s...")
                retries += 1
                time.sleep(5)
                continue
            script_status = None
            if not wait:
                return
            # Wait for the script to complete
            loops = 0
            while loops < timeout and script_status not in (1, 2):
                loops += 1
                try:
                    script_status = self._device.chdkScriptStatus(script_id)
                except PTPError:
                    pass
                time.sleep(0.01)
            if not script_status:
                self.logger.warn("Script timed out, retrying...")
                retries += 1
            else:
                break
        if get_result:
            return self._get_messages(script_id)
        else:
            self._flush_messages()
            return

    def _get_messages(self, script_id):
        msg = None
        retvals = []
        # Get all messages returned by the script
        while not retvals or msg[2] != 0:
            time.sleep(0.01)
            try:
                msg = self._device.chdkReadScriptMessage()
            except PTPError:
                self.logger.warn("Couldn't read message, retrying...")
                continue
            if msg[1] == 1:
                raise DeviceException("Lua error: {0}".format(msg[0]))
            if msg[2] == 0:
                continue
            if msg[2] != script_id:
                self.logger.warn(
                    "Script IDs did not match. Expected \"{0}\","
                    " got \"{1}\", ignoring".format(script_id, msg[2])
                )
                self.logger.debug("Message (type {0}) was: {1}"
                                  .format(msg[1], msg[0]))
                continue
            self.logger.debug("Camera returned: {0}".format(msg[0]))
            retvals.append(msg[0])
        # NOTE: Just to be safe...
        time.sleep(0.25)
        # Return a tuple containing all of the messages
        return tuple(retvals)

    def _flush_messages(self):
        msg = (None, None, None)
        tries = 0
        while msg[2] != 0:
            tries += 1
            time.sleep(0.01)
            try:
                msg = self._device.chdkReadScriptMessage()
            except PTPError:
                if tries > 10:
                    return
                self.logger.warn("Couldn't read message, ignoring...")

    def get_orientation(self):
        # NOTE: CHDK doesn't seem to be able to write the EXIF owner to the
        #       camera's internal flash memory, so we have to use this
        #       workaround.
        if self._orientation:
            return self._orientation
        try:
            orientation = self.execute_lua("""
                file=io.open('A/OWN.TXT')
                content=file:read()
                file:close()
                return content
                """, get_result=True)[0].lower()
            self.logger.debug("Orientation is: \"{0}\"".format(orientation))
            self.logger = logging.getLogger('PTPDevice[{0}]'
                                            .format(orientation))
            self._orientation = orientation
            return orientation
        except DeviceException as e:
            self.logger.warn("Could not get orientation, reason: {0}"
                             .format(e.message))
            return None

    def set_orientation(self, orientation):
        self.execute_lua("""
            file=io.open('A/OWN.TXT', 'w')
            file:write("{0}")
            file:close()
        """.format(orientation.upper()), get_result=False)
        self._orientation = orientation

    def get_image_list(self):
        img_path = self.execute_lua("get_image_dir()", get_result=True)[0]
        file_list = [os.path.join(img_path, x.split("\t")[1])
                     for x in (self.execute_lua("os.listdir(\"{0}\")"
                               .format(img_path), get_result=True)[0]
                               .split("\n")[:-1])]
        return file_list

    def download_image(self, camera_path, local_path):
        self.logger.debug("Downloading \"{0}\"".format(local_path))
        self._device.chdkDownload(camera_path, local_path)
        self.logger.debug("Download complete")
        img = pexif.JpegFile.fromFile(local_path)
        if self._orientation == 'left':
            exif_orientation = 8  # 90°
        else:
            exif_orientation = 6  # -90°
        img.exif.primary.Orientation = [exif_orientation]
        img.writeFile(local_path)

    def delete_image(self, camera_path):
        self.execute_lua("os.remove(\"{0}\")".format(camera_path))
        time.sleep(0.5)

    def download_all_images(self, path):
        for fpath in self.get_image_list():
            self.download_image(
                fpath,
                os.path.join(path, os.path.basename(fpath)))

    def delete_all_images(self):
        for fpath in self.get_image_list():
            self.delete_image(fpath)

    def get_preview_image(self, viewport=True, ui_overlay=False):
        flags = 0
        if viewport:
            flags |= pyptpchdk.LiveViewFlags.VIEWPORT
        if ui_overlay:
            flags |= pyptpchdk.LiveViewFlags.BITMAP
        data = self._device.chdkGetLiveData(flags)
        img = data['viewport']
        # FIXME: Use wand for this, once the new version is out that supports
        #        raw RGB data.
        return Image.fromstring('RGB', (img['width'], img['height']),
                                img['data'])


class CHDKCameraDevice(DevicePlugin):
    """ Plugin for digital cameras running the CHDK firmware.

    """
    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument(
            '--sensitivity', '-S', dest="sensitivity", type=int,
            metavar="<int>", help="ISO sensitivity")
        parser.add_argument(
            "--shutter-speed", '-s', dest="shutter_speed", type=unicode,
            metavar="<int/float/str>", help="Shutter speed")
        parser.add_argument(
            "--zoom-level", "-z", dest="zoom_level", type=int, metavar="<int>",
            help="Zoom level")

    @classmethod
    def match(cls, device):
        # TODO: This matches all PTP devices, we will have to ensure that
        #       it is also a CHDK compatible device.
        cfg = device.get_active_configuration()[(0, 0)]
        matches = (hex(cfg.bInterfaceClass) == "0x6"
                   and hex(device.bInterfaceSubClass) == "0x1")
        return matches

    def __init__(self, config, device):
        """ Set connection information and try to obtain orientation.

        :param config:  spreads configuration
        :type config:   spreads.confit.ConfigView
        :param device:  USB device to use for the object
        :type device:   `usb.core.Device <http://github.com/walac/pyusb>`_

        """
        self._device = PTPDevice(device)
        self.orientation = self._device.get_orientation()
        self._sensitivity = config['device']['sensitivity'].get(int)
        self._shutter_speed = (float(Fraction(config['device']['shutter_speed']
                                              .get(unicode))))
        self._zoom_level = config['device']['zoom_level'].get(int)
        self._dpi = config['device']['dpi'].get(int)

    def set_orientation(self, orientation):
        """ Set the device orientation.

        :param orientation: The orientation name
        :type orientation:  unicode in (u"left", u"right")

        """
        self._device.set_orientation(orientation)
        self.orientation = orientation

    def delete_files(self):
        self._device.delete_all_images()

    def download_files(self, path):
        self._device.download_all_images(path)

    def list_files(self):
        return self._device.get_image_list()

    def prepare_capture(self):
        self._set_record_mode()
        time.sleep(0.25)
        self._set_zoom(self._zoom_level)
        self._disable_flash()
        self._disable_ndfilter()
        self._set_sensitivity(self._sensitivity)
        self._set_shutter_speed(self._shutter_speed)
        self._lock_focus()
        time.sleep(3)

    def capture(self):
        """ Shoot a picture. """
        self._set_sensitivity(self._sensitivity)
        time.sleep(0.25)
        self._set_shutter_speed(self._shutter_speed)
        time.sleep(0.25)
        self._device.execute_lua("shoot()")
        # NOTE: This is a safety measure, as CHDK tends to crash if we don't
        #       wait.
        time.sleep(1)

    def _set_record_mode(self):
        """ Put the camera in record mode. """
        self._device.execute_lua("switch_mode_usb(1)")

    def _get_zoom(self):
        """ Get current zoom level.

        :returns: int -- the current zoom level

        """
        return self._device.execute_lua("get_zoom()", get_result=True)[0]

    def _set_zoom(self, level):
        """ Set zoom level.

        :param level: The zoom level to be used
        :type level:  int

        """
        available_levels = self._device.execute_lua("get_zoom_steps()",
                                                    get_result=True)[0]
        if not level in available_levels:
            raise DeviceException("Level outside of supported range!")
        self._device.execute_lua('set_zoom({0})'.format(level))

    def _disable_flash(self):
        """ Disable the camera's flash. """
        self._device.execute_lua("set_prop(16, 2)")

    def _set_sensitivity(self, value=80):
        """ Set the camera's ISO value.

        :param iso_value: The ISO value in ISO/ASA format
        :type iso-value:  int

        """
        self._device.execute_lua("set_iso_mode({0})".format(value))

    def _set_shutter_speed(self, value=448):
        # Calculate TV96 value from shutter_speed
        tv96_value = -96*log(value)/log(2)
        # Round to nearest multiple of 32
        tv96_value = int(32*round(tv96_value/32))
        self._device.execute_lua("set_tv96_direct({0})".format(tv96_value))

    def _disable_ndfilter(self):
        """ Disable the camera's ND Filter. """
        self._device.execute_lua("set_nd_filter(2)")

    def _lock_focus(self):
        """ Acquire auto focus and lock it. """
        self._device.execute_lua("set_aflock(0)")
        self._device.execute_lua("press('shoot_half')")
        time.sleep(0.8)
        self._device.execute_lua("release('shoot_half')")
        time.sleep(0.5)
        self._device.execute_lua("set_aflock(1)")

    def _play_sound(self, sound_num):
        """ Plays one of the following sounds:
                0 = startup sound
                1 = shutter sound
                2 = button press sound
                3 = selftimer
                4 = short beep
                5 = af (auto focus) confirmation
                6 = error beep
        """
        self._device.execute_lua("play_sound({0})".format(sound_num))


class CanonA2200CameraDevice(CHDKCameraDevice):
    """ Canon A2200 driver.

        Works around some quirks of that CHDK port.

    """
    ISO_TO_APEX = {
        80: 373,
    }

    def __init__(self, config, device):
        super(CanonA2200CameraDevice, self).__init__(config, device)
        self.logger = logging.getLogger(
            'CanonA2200CameraDevice[{0}]'.format(self.orientation))

    @classmethod
    def match(cls, device):
        matches = (hex(device.idVendor) == "0x4a9"
                   and hex(device.idProduct) == "0x322a")
        return matches

    def _get_image_list(self):
        # NOTE:
        #   We use this workaround, as the A2200 doesn't have enough memory
        #   to deal with directory listings containing more than ~250
        #   items.
        #   We make use of the fact that the firmware numbers images
        #   consecutively: By obtaining the name of the first and the
        #   name of the last image, we can safely determine the rest
        #   of the filenames without getting the whole directory listing.
        img_path = self._device.execute_lua("get_image_dir()",
                                            get_result=True)[0]
        first_file = self._device.execute_lua('flist = os.listdir("{0}")\n'
                                              'table.sort(flist)\n'
                                              'return flist[1]'
                                              .format(img_path),
                                              get_result=True)[0]
        if not first_file:
            self.logger.warn("Could not get name of first file!")
            raise StopIteration
        first_num = int(first_file[4:-4])
        #last_num = self._device.execute_lua("get_exp_count()",
        #                                    get_result=True)[0]
        num_files = self._device.execute_lua("table.getn(os.listdir('{0}'))"
                                             .format(img_path), get_result=True
                                             )[0]
        last_file = self._device.execute_lua('flist = os.listdir("{0}")\n'
                                             'table.sort(flist)\n'
                                             'return flist[{1}]'
                                             .format(img_path, num_files),
                                             get_result=True)[0]
        last_num = int(last_file[4:-4])
        for num in xrange(first_num, last_num+1):
            fname = "IMG_{0}.JPG".format(num)
            yield os.path.join(img_path, fname)

    def download_files(self, path):
        for campath in self._get_image_list():
            local_path = os.path.join(path, os.path.basename(campath))
            try:
                self._device.download_image(campath, local_path)
            except:
                continue

    def delete_files(self):
        for campath in self._get_image_list():
            try:
                self._device.delete_image(campath)
            except:
                continue

    def list_files(self):
        return list(self._get_image_list())

    def _set_zoom(self, level):
        """ Set zoom level.

            The A2200 currently has a bug, where setting the zoom level
            directly via set_zoom crashes the camera quite frequently, so
            we work around that by simulating button presses.

        :param level: The zoom level to be used
        :type level:  int

        """
        if level > 7:
            raise Exception("Zoom level {0} exceeds the camera's range!"
                            .format(level))
        zoom = None
        while zoom != level:
            zoom = self._get_zoom()
            self.logger.debug("Zoom for {0}: {1}".format(self.orientation,
                                                         zoom))
            if zoom > level:
                self._device.execute_lua('click("zoom_out")')
            elif zoom < level:
                self._device.execute_lua('click("zoom_in")')
            time.sleep(1)

    def _set_sensitivity(self, value=80):
        """ Set the camera's ISO value.

            `set_iso_mode` doesn't seem to work, so we have to use set_sv96.
            Problem is that I've not yet figured out the conversion formula
            from ISO to the ASA/APEX value required for this function, which,
            from what I understand, has to be determined for every camera
            separately...

        :param iso_value: The ISO value in ISO/ASA format
        :type iso-value:  int

        """
        iso_value = self._sensitivity
        try:
            sv96_value = CanonA2200CameraDevice.ISO_TO_APEX[iso_value]
        except KeyError:
            raise Exception("The desired ISO value is not supported.")
        self._device.execute_lua("set_sv96({0})".format(sv96_value))



# -*- coding: utf-8 -*-

from __future__ import division, unicode_literals

import logging
import os
import subprocess

from spreads.plugin import HookPlugin
from spreads.util import MissingDependencyException, find_in_path

if not find_in_path('djvubind'):
    raise MissingDependencyException("Could not find executable `djvubind` in"
                                     " $PATH. Please install the appropriate"
                                     " package(s)!")

logger = logging.getLogger('spreadsplug.djvubind')


class DjvuBindPlugin(HookPlugin):
    def output(self, path):
        logger.info("Assembling DJVU.")
        img_dir = os.path.join(path, 'done')
        djvu_file = os.path.join(path, 'out',
                                 "{0}.djvu".format(os.path.basename(path)))
        cmd = ["djvubind", img_dir]
        if self.config['djvubind']['ocr'].get(unicode) == 'none':
            cmd.append("--no-ocr")
        logger.debug("Running " + " ".join(cmd))
        _ = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        os.rename("book.djvu", djvu_file)



import logging
import os
import re
import subprocess
import xml.etree.cElementTree as ET

from concurrent import futures

from spreads.plugin import HookPlugin
from spreads.util import find_in_path, MissingDependencyException

if not find_in_path('tesseract'):
    raise MissingDependencyException("Could not find executable `tesseract`"
                                     " in $PATH. Please install the"
                                     " appropriate package(s)!")

logger = logging.getLogger('spreadsplug.tesseract')


class TesseractPlugin(HookPlugin):
    @classmethod
    def add_arguments(cls, command, parser):
        if command == 'postprocess':
            parser.add_argument("--language", "-l",
                                dest="language", default="eng",
                                help="OCR language (3-letter language code)"
                                     " [default: eng]")

    def process(self, path):
        ocr_lang = self.config['language'].get(str)
        logger.info("Performing OCR")
        logger.info("Language is \"{0}\"".format(ocr_lang))
        img_dir = os.path.join(path, 'done')
        with futures.ProcessPoolExecutor() as executor:
            for img in os.listdir(img_dir):
                if not img.endswith('tif'):
                    continue
                executor.submit(
                    subprocess.check_call,
                    ["tesseract", os.path.join(img_dir, img),
                     os.path.join(img_dir, os.path.splitext(img)[0]),
                     "-l", ocr_lang, "hocr"],
                    stderr=subprocess.STDOUT
                )
        # NOTE: This modifies the hOCR files to make them compatible with
        #       pdfbeads.
        #       See the following bugreport for more information:
        #       http://rubyforge.org/tracker/index.php?func=detail&aid=29737&group_id=9752&atid=37737
        for hocr in os.listdir(img_dir):
            if not hocr.endswith('html'):
                continue
            fname = os.path.join(img_dir, hocr)
            with open(fname, 'r') as fp:
                new_content = re.sub(r'(<span[^>]*>(<strong>)? +(<\/strong>)?'
                                     r'<\/span>*)(<span[^>]*>(<strong>)? '
                                     r'+(<\/strong>)?<\/span> *)',
                                     r'\g<1>', fp.read())
            with open(fname, 'w') as fp:
                fp.write(new_content)

    def output(self, path):
        outfile = os.path.join(path, 'out', "{0}.hocr"
                                            .format(os.path.basename(path)))
        path = os.path.join(path, 'done')
        out_root = ET.Element('html')
        ET.SubElement(out_root, 'head')
        body = ET.SubElement(out_root, 'body')

        for idx, page in enumerate(sorted([os.path.join(path, x)
                                   for x in os.listdir(path)
                                   if x.endswith('.html')])):
            # NOTE: Fixe some things from hOCR output so we can parse it...
            with open(page, 'r+') as fp:
                content = re.sub(r'<em><\/em>', '', fp.read())
                content = re.sub(r'<strong><\/strong>', '', content)
            page = ET.fromstring(content)
            page_elem = page.find("xhtml:body/xhtml:div[@class='ocr_page']",
                                  dict(xhtml="http://www.w3.org/1999/xhtml"))
            # Correct page_number
            page_elem.set('id', 'page_{0}'.format(idx))

            # And tack it onto the output file
            if page_elem:
                body.append(page_elem)
        with open(outfile, 'w') as fp:
            # Strip those annoying namespace tags...
            out_string = re.sub(r"(<\/*)html:", "\g<1>",
                                ET.tostring(out_root))
            fp.write(out_string.encode("UTF-8"))



# -*- coding: utf-8 -*-

# Copyright (c) 2013 Johannes Baiter. All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

""" Various utility functions.
"""

from __future__ import division, unicode_literals

import abc
import itertools
import os
from logging import StreamHandler

from colorama import Fore, Back, Style


class SpreadsException(Exception):
    pass


class DeviceException(SpreadsException):
    pass


class MissingDependencyException(SpreadsException):
    pass


def find_in_path(name):
    """ Find executable in $PATH.

    :param name:  name of the executable
    :type name:   unicode
    :returns:     bool -- True if *name* is found or False

    """
    return name in itertools.chain(*tuple(os.listdir(x)
                                   for x in os.environ.get('PATH').split(':')
                                   if os.path.exists(x)))


class _instancemethodwrapper(object):
    def __init__(self, callable):
        self.callable = callable
        self.__dontcall__ = False

    def __getattr__(self, key):
        return getattr(self.callable, key)

    def __call__(self, *args, **kwargs):
        if self.__dontcall__:
            raise TypeError('Attempted to call abstract method.')
        return self.callable(*args, **kwargs)


class _classmethod(classmethod):
    def __init__(self, func):
        super(_classmethod, self).__init__(func)
        isabstractmethod = getattr(func, '__isabstractmethod__', False)
        if isabstractmethod:
            self.__isabstractmethod__ = isabstractmethod

    def __get__(self, instance, owner):
        result = _instancemethodwrapper(super(_classmethod, self)
                                        .__get__(instance, owner))
        isabstractmethod = getattr(self, '__isabstractmethod__', False)
        if isabstractmethod:
            result.__isabstractmethod__ = isabstractmethod
            abstractmethods = getattr(owner, '__abstractmethods__', None)
            if abstractmethods and result.__name__ in abstractmethods:
                result.__dontcall__ = True
        return result


class abstractclassmethod(_classmethod):
    """ New decorator class that implements the @abstractclassmethod decorator
        added in Python 3.3 for Python 2.7.

        Kudos to http://stackoverflow.com/a/13640018/487903

    """
    def __init__(self, func):
        func = abc.abstractmethod(func)
        super(abstractclassmethod, self).__init__(func)


class ColourStreamHandler(StreamHandler):
    """ A colorized output SteamHandler
    Kudos to Leigh MacDonald:
    http://leigh.cudd.li/article/Cross_Platform_Colorized_Logger_Output_Using_Pythons_logging_Module_And_Colorama
    """

    # Some basic colour scheme defaults
    colours = {
        'DEBUG': Fore.CYAN,
        'INFO': Fore.GREEN,
        'WARN': Fore.YELLOW,
        'WARNING': Fore.YELLOW,
        'ERROR': Fore.RED,
        'CRIT': Back.RED + Fore.WHITE,
        'CRITICAL': Back.RED + Fore.WHITE
    }

    @property
    def is_tty(self):
        """ Check if we are using a "real" TTY. If we are not using a TTY it
        means that the colour output should be disabled.

        :return: Using a TTY status
        :rtype: bool
        """
        try:
            return getattr(self.stream, 'isatty', None)()
        except:
            return False

    def emit(self, record):
        try:
            message = self.format(record)
            if not self.is_tty:
                self.stream.write(message)
            else:
                self.stream.write(self.colours[record.levelname]
                                  + message + Style.RESET_ALL)
            self.stream.write(getattr(self, 'terminator', '\n'))
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)



import logging
import os
import time

import subprocess
from concurrent.futures import ThreadPoolExecutor
from PySide import QtCore, QtGui
from PIL import Image

import spreads
import spreads.workflow as workflow
import spreads.util as util
from spreads.plugin import get_devices, get_pluginmanager

import gui_rc


class LogBoxHandler(logging.Handler):
    # NOTE: This is neccessary, because a signal has to be an attribute of
    #       a QObject instance. Multiple inheritance does not work here,
    #       as both logging.Handler and QObject have an "emit" method
    class DummyQObject(QtCore.QObject):
        sig = QtCore.Signal(unicode)

    def __init__(self, textedit):
        logging.Handler.__init__(self)
        self._qobj = self.DummyQObject()
        self.sig = self._qobj.sig
        self.textedit = textedit

    def emit(self, record):
        # Play a warning sound if something has gone wrong
        self.sig.emit(self.format(record))
        if record.levelname in ('ERROR', 'CRITICAL', 'WARNING'):
            QtGui.QApplication.beep()


class LogBoxFormatter(logging.Formatter):
    LEVELS = {
        'ERROR': "<strong><font color=\"red\">{0}</font></strong>",
        'CRITICAL': "<strong><font color=\"orange\">{0}</font></strong>",
        'WARNING': "<strong><font color=\"orange\">{0}</font></strong>",
        'INFO': "{0}",
        'DEBUG': "{0}",
    }

    def format(self, record):
        levelname = record.levelname
        return self.LEVELS[levelname].format(record.message)


class SpreadsWizard(QtGui.QWizard):
    def __init__(self, config, parent=None):
        super(SpreadsWizard, self).__init__(parent)
        self.addPage(IntroPage())
        self.addPage(ConnectPage())
        self.addPage(CapturePage())
        self.addPage(DownloadPage())
        self.addPage(PostprocessPage())
        self.addPage(OutputPage())
        self.addPage(FinishPage())

        self.setPixmap(QtGui.QWizard.WatermarkPixmap,
                       QtGui.QPixmap(':/pixmaps/monk.png'))

        button_layout = []
        button_layout.append(QtGui.QWizard.BackButton)
        button_layout.append(QtGui.QWizard.Stretch)
        button_layout.append(QtGui.QWizard.CancelButton)
        button_layout.append(QtGui.QWizard.Stretch)
        button_layout.append(QtGui.QWizard.NextButton)
        button_layout.append(QtGui.QWizard.FinishButton)
        self.setButtonLayout(button_layout)

        self.setWindowTitle("Spreads Wizard")


class IntroPage(QtGui.QWizardPage):
    def initializePage(self):
        self.wizard().active_plugins = get_pluginmanager().names()
        self.setTitle("Welcome!")

        intro_label = QtGui.QLabel(
            "This wizard will guide you through the digitization workflow. "
        )

        dirpick_layout = QtGui.QHBoxLayout()
        self.line_edit = QtGui.QLineEdit()
        browse_btn = QtGui.QPushButton("Browse")
        dirpick_layout.addWidget(self.line_edit)
        dirpick_layout.addWidget(browse_btn)
        browse_btn.clicked.connect(self.show_filepicker)

        form_layout = QtGui.QFormLayout()

        self.keep_box = QtGui.QCheckBox("Keep files on devices")
        form_layout.addRow(self.keep_box)

        self.even_device = None
        if ('combine' in self.wizard().active_plugins
                or 'autorotate' in self.wizard().active_plugins):
            self.even_device = QtGui.QComboBox()
            self.even_device.addItem("Left")
            self.even_device.addItem("Right")
            self.even_device.setCurrentIndex(0)
            form_layout.addRow("Device for even pages", self.even_device)

        self.st_auto = None
        self.st_detection = None
        if 'scantailor' in self.wizard().active_plugins:
            self.st_auto = QtGui.QCheckBox("Don't Inspect ScanTailor"
                                           " configuration")
            form_layout.addRow(self.st_auto)
            self.st_detection = QtGui.QComboBox()
            self.st_detection.addItem("Content")
            self.st_detection.addItem("Page")
            self.st_detection.setCurrentIndex(0)
            form_layout.addRow("ScanTailor layout mode", self.st_detection)

        self.ocr_lang = None
        if 'tesseract' in self.wizard().active_plugins:
            langs = (subprocess.check_output(["tesseract", "--list-langs"],
                                             stderr=subprocess.STDOUT)
                     .split("\n")[1:-1])
            self.ocr_lang = QtGui.QComboBox()
            for lang in langs:
                self.ocr_lang.addItem(lang)
            self.ocr_lang.setCurrentIndex(0)
            form_layout.addRow("OCR language", self.ocr_lang)

        main_layout = QtGui.QVBoxLayout()
        main_layout.addWidget(intro_label)
        main_layout.addSpacing(30)
        main_layout.addWidget(
            QtGui.QLabel("Please select a project directory.")
        )
        main_layout.addLayout(dirpick_layout)
        main_layout.addSpacing(30)
        main_layout.addLayout(form_layout)

        self.setLayout(main_layout)

    def show_filepicker(self):
        dialog = QtGui.QFileDialog()
        dialog.setFileMode(QtGui.QFileDialog.Directory)
        dialog.setOption(QtGui.QFileDialog.ShowDirsOnly, True)
        dialog.exec_()
        self.line_edit.setText(dialog.selectedFiles()[0])

    def validatePage(self):
        self.wizard().project_path = self.line_edit.text()
        if not self.wizard().project_path:
            msg_box = QtGui.QMessageBox()
            msg_box.setText("Please select a project directory.")
            msg_box.setIcon(QtGui.QMessageBox.Critical)
            msg_box.exec_()
            return False

        spreads.config['keep'] = self.keep_box.isChecked()
        if self.even_device and self.even_device.currentText() != 'Left':
            spreads.config['first_page'] = "right"
            spreads.config['rotate_inverse'] = True
        if self.st_auto:
            spreads.config['autopilot'] = self.st_auto.isChecked()
        if self.st_detection:
            spreads.config['page_detection'] = (
                self.st_detection.currentText().lower() == 'page')
        if self.ocr_lang:
            spreads.config['language'] = str(self.ocr_lang.currentText()
                                             .lower())
        return True


class ConnectPage(QtGui.QWizardPage):
    def initializePage(self):
        # TODO: Show an activity indicator while get_devices() is running,
        #       use QtGui.QProgressDialog, with .setRange(0,0) for
        #       indeterminate time

        self.setTitle("Connect")

        self.label = QtGui.QLabel("Detecting devices...")
        self.devicewidget = QtGui.QListWidget()
        self.devicewidget.setSelectionMode(QtGui.QAbstractItemView
                                           .MultiSelection)

        refresh_btn = QtGui.QPushButton("Refresh")
        refresh_btn.clicked.connect(self.update_devices)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.devicewidget)
        layout.addWidget(refresh_btn)
        self.setLayout(layout)
        self.update_devices()

    def update_devices(self):
        progress = QtGui.QProgressDialog("Detecting devices...",
                                         "Cancel", 0, 0, self)
        progress.show()
        self.devicewidget.clear()
        self.label.setText("Detecting devices...")
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(get_devices)
            while not future.done():
                QtGui.qApp.processEvents()
                progress.setValue(1)
                time.sleep(0.001)
            try:
                devices = future.result()
            except util.DeviceException:
                devices = []
        progress.close()
        if not devices:
            self.label.setText("<font color=red>No devices found!</font>")
        else:
            self.label.setText("Please select one or more devices:")
        self.devices = []
        for idx, device in enumerate(devices, 1000):
            self.devicewidget.setCurrentItem(
                QtGui.QListWidgetItem(device.__class__.__name__,
                                      self.devicewidget, type=idx))
            self.devices.append(device)

    def validatePage(self):
        self.wizard().selected_devices = [self.devices[x.type()-1000]
                                          for x
                                          in self.devicewidget.selectedItems()]

        if len(self.wizard().selected_devices) in (1, 2):
            progress = QtGui.QProgressDialog("Preparing devices...",
                                             "Cancel", 0, 0, self)
            progress.show()
            with ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(workflow.prepare_capture,
                                         self.wizard().selected_devices)
                while future.running():
                    QtGui.qApp.processEvents()
                    progress.setValue(1)
                    time.sleep(0.001)
            progress.close()
            return True
        msg_box = QtGui.QMessageBox()
        msg_box.setIcon(QtGui.QMessageBox.Critical)
        if len(self.wizard().selected_devices) > 2:
            msg_box.setText("Please don't select more than two devices!")
        elif not self.wizard().selected_devices:
            msg_box.setText("No device selected.")
        msg_box.exec_()
        return False


class CapturePage(QtGui.QWizardPage):
    def initializePage(self):
        self.setTitle("Capturing from devices")
        self.start_time = None
        self.shot_count = None

        layout = QtGui.QVBoxLayout(self)
        self.status = QtGui.QLabel("Press a capture key (default: Space, B)"
                                   " to begin capturing.")

        previewbox = QtGui.QHBoxLayout()
        self.left_preview = QtGui.QLabel()
        self.right_preview = QtGui.QLabel()
        previewbox.addWidget(self.left_preview)
        previewbox.addWidget(self.right_preview)

        self.refresh_btn = QtGui.QPushButton("Refresh")
        self.refresh_btn.clicked.connect(self.update_preview)

        self.capture_btn = QtGui.QPushButton("Capture")
        self.capture_btn.clicked.connect(self.doCapture)
        self.capture_btn.setFocus()

        self.logbox = QtGui.QTextEdit()
        self.log_handler = LogBoxHandler(self.logbox)
        self.log_handler.setLevel(logging.WARNING)
        self.log_handler.setFormatter(LogBoxFormatter())
        logging.getLogger().addHandler(self.log_handler)
        self.log_handler.sig.connect(self.logbox.append)

        layout.addWidget(self.status)
        layout.addWidget(self.refresh_btn)
        layout.addLayout(previewbox)
        layout.addWidget(self.capture_btn)
        layout.addWidget(self.logbox)
        self.setLayout(layout)
        time.sleep(0.5)
        self.update_preview()

    def validatePage(self):
        workflow.finish_capture(self.wizard().selected_devices)
        return True

    def keyPressEvent(self, event):
        if event.key() in (QtCore.Qt.Key_B, QtCore.Qt.Key_Space):
            self.doCapture()

    def doCapture(self):
        if self.start_time is None:
            self.start_time = time.time()
        if self.shot_count is None:
            self.shot_count = 0
        self.capture_btn.setEnabled(False)
        self.refresh_btn.setEnabled(False)
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(workflow.capture,
                                     self.wizard().selected_devices)
            while future.running():
                QtGui.qApp.processEvents()
                time.sleep(0.001)
        self.capture_btn.setEnabled(True)
        self.refresh_btn.setEnabled(True)
        self.capture_btn.setFocus()
        self.shot_count += 2
        self.status.setText("Shot {0} pages in {1:.0f} minutes "
                            "({2:.0f} pages/hour)".format(
                                self.shot_count,
                                (time.time() - self.start_time) / 60,
                                ((3600 / (time.time() - self.start_time))
                                 * self.shot_count)))

    def update_preview(self):
        def get_preview(dev):
            img = dev._device.get_preview_image()
            img = img.resize((320, 240), Image.ANTIALIAS).convert('RGBA')
            data = img.tostring('raw')
            image = QtGui.QImage(data, img.size[0], img.size[1],
                                 QtGui.QImage.Format_ARGB32).rgbSwapped()
            return dev.orientation, image

        # TODO: Don't go via PIL, find a way to use RGB data directly
        with ThreadPoolExecutor(max_workers=2) as executor:
            futures = executor.map(get_preview,
                                   self.wizard().selected_devices)
            previews = tuple(futures)
        for orientation, image in previews:
            pixmap = QtGui.QPixmap.fromImage(image)
            if orientation == 'left':
                self.left_preview.setPixmap(pixmap)
            else:
                self.right_preview.setPixmap(pixmap)


class DownloadPage(QtGui.QWizardPage):
    def initializePage(self):
        self.setTitle("Downloading")

        #TODO: Display a real progress bar here, update it according to the
        #      combined number of files in raw or its subfolders
        self.progressbar = QtGui.QProgressBar(self)
        self.progressbar.setRange(0, 0)
        self.progressbar.setAlignment(QtCore.Qt.AlignCenter)
        self.logbox = QtGui.QTextEdit()
        self.log_handler = LogBoxHandler(self.logbox)
        self.log_handler.setLevel(logging.INFO)
        self.log_handler.setFormatter(LogBoxFormatter())
        logging.getLogger().addHandler(self.log_handler)
        self.log_handler.sig.connect(self.logbox.append)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.progressbar)
        layout.addWidget(self.logbox)
        self.setLayout(layout)
        self.logbox.clear()
        QtCore.QTimer.singleShot(0, self.doDownload)

    def doDownload(self):
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(workflow.download,
                                     self.wizard().selected_devices,
                                     self.wizard().project_path)
            while future.running():
                QtGui.qApp.processEvents()
                self.progressbar.setValue(1)
                time.sleep(0.001)
            if future.exception():
                raise future.exception()
        self.progressbar.hide()
        if ('combine' in self.wizard().active_plugins and not
                os.listdir(os.path.join(self.wizard().project_path, 'raw'))):
            msg_box = QtGui.QMessageBox()
            msg_box.setText("It seems that combining the images has failed."
                            "Please fix the issue manually and retry.")
            # TODO: Add a 'combine' button to retry combining.
            msg_box.setIcon(QtGui.QMessageBox.Critical)
            self.combine_btn = msg_box.addButton("Combine",
                                                 QtGui.QMessageBox.ApplyRole)
            self.combine_btn.clicked.connect(get_pluginmanager()['combine']
                                             .download(self.wizard()
                                             .project_path))
            msg_box.exec_()

    def validatePage(self):
        logging.getLogger().removeHandler(self.log_handler)
        return True


class PostprocessPage(QtGui.QWizardPage):
    def initializePage(self):
        self.setTitle("Postprocessing")

        self.progressbar = QtGui.QProgressBar(self)
        self.progressbar.setRange(0, 0)
        self.progressbar.setAlignment(QtCore.Qt.AlignCenter)
        self.logbox = QtGui.QTextEdit()
        self.log_handler = LogBoxHandler(self.logbox)
        self.log_handler.setLevel(logging.INFO)
        self.log_handler.setFormatter(LogBoxFormatter())
        logging.getLogger().addHandler(self.log_handler)
        self.log_handler.sig.connect(self.logbox.append)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.progressbar)
        layout.addWidget(self.logbox)
        self.setLayout(layout)
        self.logbox.clear()
        QtCore.QTimer.singleShot(0, self.doPostprocess)

    def doPostprocess(self):
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(workflow.process,
                                     self.wizard().project_path)
            while not future.done():
                QtGui.qApp.processEvents()
                self.progressbar.setValue(1)
                time.sleep(0.01)
            if future.exception():
                raise future.exception()
        self.progressbar.hide()

    def validatePage(self):
        logging.getLogger().removeHandler(self.log_handler)
        return True


class OutputPage(QtGui.QWizardPage):
    def initializePage(self):
        self.setTitle("Generating output files")

        self.progressbar = QtGui.QProgressBar(self)
        self.progressbar.setRange(0, 0)
        self.progressbar.setAlignment(QtCore.Qt.AlignCenter)
        self.logbox = QtGui.QTextEdit()
        self.log_handler = LogBoxHandler(self.logbox)
        self.log_handler.setLevel(logging.INFO)
        self.log_handler.setFormatter(LogBoxFormatter())
        logging.getLogger().addHandler(self.log_handler)
        self.log_handler.sig.connect(self.logbox.append)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.progressbar)
        layout.addWidget(self.logbox)
        self.setLayout(layout)
        self.logbox.clear()
        QtCore.QTimer.singleShot(0, self.doGenerateOutput)

    def doGenerateOutput(self):
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(workflow.output,
                                     self.wizard().project_path)
            while not future.done():
                QtGui.qApp.processEvents()
                self.progressbar.setValue(1)
                time.sleep(0.001)
        self.progressbar.hide()

    def validatePage(self):
        logging.getLogger().removeHandler(self.log_handler)
        return True


class FinishPage(QtGui.QWizardPage):
    def initializePage(self):
        self.setFinalPage(True)
        self.setTitle("Done!")
        # TODO: Offer option to view project folder on exit



# -*- coding: utf-8 -*-

# Copyright (c) 2013 Johannes Baiter. All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import division, unicode_literals

import abc
import logging

import usb
import stevedore
from stevedore.extension import ExtensionManager
from stevedore.named import NamedExtensionManager


import spreads
from spreads.util import (abstractclassmethod, find_in_path, SpreadsException,
                          DeviceException)

logger = logging.getLogger("spreads.plugin")


class SpreadsNamedExtensionManager(NamedExtensionManager):
    """ Custom extension manager for spreads.

    stevedore's NamedExtensionmanger does not give us the Exception that caused
    a plugin to fail at initialization. This derived class throws the original
    exception instead of logging it.
    """
    def _load_plugins(self, invoke_on_load, invoke_args, invoke_kwds):
        extensions = []
        for ep in self._find_entry_points(self.namespace):
            stevedore.LOG.debug('found extension %r', ep)
            ext = self._load_one_plugin(ep,
                                        invoke_on_load,
                                        invoke_args,
                                        invoke_kwds,
                                        )
            if ext:
                extensions.append(ext)
        return extensions


class SpreadsPlugin(object):
    """ Plugin base class.

    """

    @classmethod
    def add_arguments(cls, command, parser):
        """ Allows a plugin to register new arguments with the command-line
            parser.

        :param command: The command to be modified
        :type command: unicode
        :param parser: The parser that can be modified
        :type parser: argparse.ArgumentParser

        """
        pass

    def __init__(self, config):
        """ Initialize the plugin.

        :param config: The global configuration object
        :type config: confit.ConfigView

        """
        self.config = config


class DevicePlugin(SpreadsPlugin):
    """ Base class for devices.

        Subclass to implement support for different devices.

    """
    __metaclass__ = abc.ABCMeta

    @abstractclassmethod
    def match(cls, usbdevice):
        """ Match device against USB device information.

        :param device:      The USB device to match against
        :type vendor_id:    `usb.core.Device <http://github.com/walac/pyusb>`_
        :return:            True if the :param usbdevice: matches the
                            implemented category

        """
        raise NotImplementedError

    def __init__(self, config, device):
        """ Set connection information and other properties.

        :param config:  spreads configuration
        :type config:   spreads.confit.ConfigView
        :param device:  USB device to use for the object
        :type device:   `usb.core.Device <http://github.com/walac/pyusb>`_

        """
        super(DevicePlugin, self).__init__(config['device'])
        self._device = device

    def set_orientation(self, orientation):
        """ Set the device orientation, if applicable.

        :param orientation: The orientation name
        :type orientation:  unicode in (u"left", u"right")

        """
        raise NotImplementedError

    @abc.abstractmethod
    def list_files(self):
        """ List all files on the device.

        :return: list() - The files on the device

        """
        raise NotImplementedError

    @abc.abstractmethod
    def download_files(self, local_path):
        """ Download all files from the device.

        :param local_path:  The destination path for the downloaded files
        :type local_path:   unicode

        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_files(self):
        """ Delete all files from the device.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_capture(self):
        """ Prepare device for scanning.

            What this means exactly is up to the implementation and the type,
            of device, usually it involves things like switching into record
            mode and applying all relevant settings.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def capture(self):
        """ Capture a single image with the device.

        """
        raise NotImplementedError


class HookPlugin(SpreadsPlugin):
    """ Add functionality to any of spreads' commands by implementing one or
        more of the available hooks.

    """
    @classmethod
    def add_command_parser(cls, rootparser):
        """ Allows a plugin to register a new command with the command-line
            parser. The subparser that is added to :param rootparser: should
            set the class' ``__call__`` method as the ``func`` (via
            ``set_defaults``) that is executed when the subcommand is specified
            on the CLI.

        :param rootparser: The root parser that this plugin should add a
                           subparser to.
        :type rootparser:  argparse.ArgumentParser

        """
        pass

    def prepare_capture(self, devices):
        """ Perform some action before capturing begins.

        :param devices: The devices used for capturing
        :type devices: list(DevicePlugin)

        """
        pass

    def capture(self, devices):
        """ Perform some action after each successful capture.

        :param devices: The devices used for capturing
        :type devices: list(DevicePlugin)

        """
        pass

    def finish_capture(self, devices):
        """ Perform some action after capturing has finished.

        :param devices: The devices used for capturing
        :type devices: list(DevicePlugin)

        """
        pass

    def download(self, devices, path):
        """ Perform some action after download from devices has finished.
            
            Retains all of the original information (i.e: rotation,
            metadata annotations).

        :param devices: The devices that were downloaded from.
        :type devices: list(DevicePlugin)
        :param path: The destination directory for the downloads
        :type path: unicode

        """
        pass

    def delete(self, devices):
        """ Perform some action after images have been deleted from the
            devices.

        :param devices: The devices that were downloaded from.
        :type devices: list(DevicePlugin)

        """
        pass

    def process(self, path):
        """ Perform one or more actions that either modify the captured images
            or generate a different output.

        .. note:
            This method is intended to operate on the *done* subdfolder of
            the project directory. At the beginning of postprocessing, it
            will contain copies of the images in *raw*. This is to ensure that
            a copy of the original, scanned images will always be available
            for archival purposes.

        :param path: The project path
        :type path: unicode

        """
        pass

    def output(self, path):
        """ Assemble an output file from the postprocessed images.

        .. note:
            This method is intended to take its input files from the *done*
            subfolder of the project path and store its output in the
            *out* subfolder.

        :param path: The project path
        :type path: unicode
        """
        pass


# Load drivers for all supported devices
def get_devicemanager():
    logger.debug("Creating device manager")
    return ExtensionManager(namespace='spreadsplug.devices')

# Poor man's memoization...
_pluginmanager = None
def get_pluginmanager():
    global _pluginmanager
    if _pluginmanager:
        return _pluginmanager
    logger.debug("Creating plugin manager")
    _pluginmanager = SpreadsNamedExtensionManager(
        namespace='spreadsplug.hooks',
        names=spreads.config['plugins'].as_str_seq(),
        invoke_on_load=True,
        invoke_args=[spreads.config],
        name_order=True)
    return _pluginmanager


def get_devices():
    """ Detect all attached devices and select a fitting driver.

    :returns:  list(DevicePlugin) -- All supported devices that were detected

    """
    def match(extension, device):
        try:
            devname = usb.util.get_string(device, 256, 2)
        except:
            devname = "{0}:{1}".format(device.bus, device.address)
        logger.debug("Trying to match device \"{0}\" with plugin {1}"
                     .format(devname, extension.plugin.__name__))
        try:
            match = extension.plugin.match(device)
        # Ignore devices that don't implement `match`
        except TypeError:
            logger.debug("Plugin did not implement match method!")
            return
        if match:
            logger.debug("Plugin matched device!")
            return extension.plugin
    logger.debug("Detecting support for attached devices")
    devices = []
    candidates = usb.core.find(find_all=True)
    devicemanager = get_devicemanager()
    for device in candidates:
        matches = filter(None, devicemanager.map(match, device))

        # FIXME: Make this more robust: What if, for instance, two plugins
        #        are found for a device, one of which inherits from the other?
        if matches:
            devices.append(matches[0](spreads.config, device))
    if not devices:
        raise DeviceException("Could not find any compatible devices!")
    return devices



import logging
import operator
import os
import shutil

from spreads.plugin import HookPlugin

logger = logging.getLogger('spreadsplug.combine')


class CombinePlugin(HookPlugin):
    @classmethod
    def add_arguments(cls, command, parser):
        if command == 'download':
            parser.add_argument("--first-page", "-fp",
                                dest="first_page", default="left",
                                help="Set device with first page (left/right)"
                                " [default: left]")

    def download(self, cameras, path):
        left_dir = os.path.join(path, 'left')
        right_dir = os.path.join(path, 'right')
        target_dir = os.path.join(path, 'raw')
        if not os.path.exists(target_dir):
            os.mkdir(target_dir)
        left_pages = [os.path.join(left_dir, x)
                      for x in sorted(os.listdir(left_dir))]
        right_pages = [os.path.join(right_dir, x)
                       for x in sorted(os.listdir(right_dir))]
        # Write the orientation as a JPEG comment to the end of the file
        if len(left_pages) != len(right_pages):
            logger.warn("The left and right camera produced an inequal"
                        " amount of images, please fix the problem!")
            logger.warn("Will not combine images")
            return
        if (self.config['first_page']
                and not self.config['first_page'].get(str) == 'left'):
            combined_pages = reduce(operator.add, zip(right_pages, left_pages))
        else:
            combined_pages = reduce(operator.add, zip(left_pages, right_pages))
        logger.info("Combining images.")
        for idx, fname in enumerate(combined_pages):
            fext = os.path.splitext(os.path.split(fname)[1])[1]
            target_file = os.path.join(target_dir, "{0:04d}{1}"
                                       .format(idx, fext))
            shutil.copyfile(fname, target_file)
        shutil.rmtree(right_dir)
        shutil.rmtree(left_dir)



from itertools import chain, repeat

from mock import MagicMock as Mock
from nose.tools import assert_raises
from pyptpchdk import PTPError

import spreads
import spreadsplug.dev.chdkcamera as chdkcamera

spreads.util.find_in_path = Mock(return_value=True)


class TestPTPDevice(object):
    def setUp(self):
        spreads.config.clear()
        spreads.config.read(user=False)
        chdkcamera.pyptpchdk = Mock()
        chdkcamera.time = Mock()
        usbmock = Mock()
        usbmock.bus, usbmock.address = 1, 2
        self.dev = chdkcamera.PTPDevice(usbmock)

    def test_execute_lua(self):
        self.dev._flush_messages = Mock()
        self.dev._device.chdkScriptStatus.side_effect = [1]
        self.dev.execute_lua("foobar")
        chdkcamera.pyptpchdk.PTPDevice.assert_called_with(1, 2)
        self.dev._device.chdkExecLua.assert_called_with("foobar")

    def test_execute_lua_nowait(self):
        self.dev._flush_messages = Mock()
        self.dev._device.chdkScriptStatus.side_effect = [1]
        self.dev.execute_lua("foobar", wait=False)
        chdkcamera.pyptpchdk.PTPDevice.assert_called_with(1, 2)
        self.dev._device.chdkExecLua.assert_called_with("foobar")

    def test_execute_lua_with_intermittent_error(self):
        self.dev._flush_messages = Mock()
        self.dev._device.chdkExecLua.side_effect = [Exception, Exception, 1]
        self.dev._device.chdkScriptStatus.side_effect = [1]
        self.dev.execute_lua("foobar")
        self.dev._device.chdkExecLua.assert_called_with("foobar")

    def test_execute_lua_with_persistent_error(self):
        self.dev._flush_messages = Mock()
        self.dev._device.chdkExecLua.side_effect = Exception
        self.dev._device.chdkScriptStatus.side_effect = [1]
        assert_raises(spreads.util.DeviceException, self.dev.execute_lua,
                      "foobar")

    def test_execute_lua_with_intermittent_timeout(self):
        self.dev._flush_messages = Mock()
        self.dev._device.chdkExecLua.return_value = 1
        self.dev._device.chdkScriptStatus.side_effect = (
            chain((PTPError,), repeat(0, 255), (1, )))
        self.dev.execute_lua("foobar")
        self.dev._device.chdkExecLua.assert_called_with("foobar")

    def test_execute_lua_with_persistent_timeout(self):
        self.dev._flush_messages = Mock()
        self.dev._device.chdkExecLua.return_value = 1
        self.dev._device.chdkScriptStatus.return_value = 0
        assert_raises(spreads.util.DeviceException, self.dev.execute_lua,
                      "foobar")

    def test_execute_lua_with_get_result(self):
        self.dev._get_messages = Mock(side_effect=["bar"])
        self.dev._device.chdkScriptStatus.side_effect = [1]
        assert self.dev.execute_lua("foo", get_result=True) == "bar"
        chdkcamera.pyptpchdk.PTPDevice.assert_called_with(1, 2)
        self.dev._device.chdkExecLua.assert_called_with("return(foo)")

    def test_execute_lua_with_get_result_hasreturn(self):
        self.dev._get_messages = Mock()
        self.dev._device.chdkScriptStatus.side_effect = [1]
        self.dev.execute_lua("return(moo)", get_result=True)
        self.dev._device.chdkExecLua.assert_called_with("return(moo)")

    def test_get_messages(self):
        self.dev._device.chdkReadScriptMessage.side_effect = (
            [("foobar", 0, 1234), (0, 0, 0)])
        msg = self.dev._get_messages(1234)
        assert msg == ("foobar", )

    def test_get_messages_readerror(self):
        self.dev._device.chdkReadScriptMessage.side_effect = (
            [PTPError, ("foobar", 0, 1234), (0, 0, 0)])
        assert self.dev._get_messages(1234) == ("foobar", )

    def test_get_messages_luaerror(self):
        self.dev._device.chdkReadScriptMessage.side_effect = (
            [("foobar", 1, 1234), (0, 0, 0)])
        assert_raises(spreads.util.DeviceException, self.dev._get_messages, 1234)

    def test_get_messages_idmismatch(self):
        self.dev._device.chdkReadScriptMessage.side_effect = (
            [("foobar", 2, 4321), ("foobar", 2, 1234), (0, 0, 0)])
        msg = self.dev._get_messages(1234)
        assert msg == ("foobar", )

    def test_flush_messages(self):
        self.dev._device.chdkReadScriptMessage.side_effect = [
            (0, 0, x) for x in reversed(xrange(10))]
        self.dev._flush_messages()
        assert self.dev._device.chdkReadScriptMessage.call_count == 10

    def test_flush_messages_error(self):
        self.dev._device.chdkReadScriptMessage.side_effect = repeat(PTPError)
        assert self.dev._flush_messages() is None
        assert self.dev._device.chdkReadScriptMessage.call_count == 11

    def test_get_orientation(self):
        self.dev._get_messages = Mock(return_value=['LEFT'])
        assert self.dev.get_orientation() == 'left'
        # Test caching
        assert self.dev.get_orientation() == 'left'
        assert self.dev._get_messages.call_count == 1

    def test_get_orientation_error(self):
        self.dev._get_messages = Mock(
            side_effect=[spreads.util.DeviceException])
        assert self.dev.get_orientation() == None

    def test_set_orientation(self):
        self.dev._flush_messages = Mock()
        self.dev.set_orientation('left')
        assert self.dev._orientation == 'left'

    def test_get_image_list(self):
        # TODO: Get camera output to mock the execute_lua call
        pass

    def test_download_image(self):
        self.dev._device.chdkDownload = Mock()
        mock_img = Mock()
        chdkcamera.pexif.JpegFile.fromFile = Mock(return_value=mock_img)
        self.dev._orientation = 'right';
        self.dev.download_image('/foo/bar.JPG', '/tmp/moo.JPG')
        assert mock_img.exif.primary.Orientation == [6]
        self.dev._orientation = 'left';
        self.dev.download_image('/foo/bar.JPG', '/tmp/moo.JPG')
        assert mock_img.exif.primary.Orientation == [8]

    def test_download_all_images(self):
        self.dev.get_image_list = Mock(return_value=repeat('foo.jpg', 10))
        self.dev.download_image = Mock()
        self.dev.download_all_images('/tmp')
        assert self.dev.download_image.call_count == 10

    def test_delete_image(self):
        # Not much to test here, but let's do it for our coverage vanity...
        self.dev.execute_lua = Mock()
        self.dev.delete_image('/a/b.jpg')

    def test_delete_all_images(self):
        self.dev.get_image_list = Mock(return_value=repeat('foo.jpg', 10))
        self.dev.delete_image = Mock()
        self.dev.delete_all_images()
        assert self.dev.delete_image.call_count == 10

    def test_get_preview_image(self):
        # Pure vanity....
        chdkcamera.Image = Mock()
        chdkcamera.pyptpchdk.LiveViewFlags.VIEWPORT = 1
        chdkcamera.pyptpchdk.LiveViewFlags.BITMAP = 2
        self.dev.get_preview_image(viewport=True, ui_overlay=True)
        self.dev._device.chdkGetLiveData.assert_called_with(3)



# -*- coding: utf-8 -*-

from __future__ import division, unicode_literals

import logging
import math
import multiprocessing
import os
import shutil
import subprocess
import tempfile
from xml.etree.cElementTree import ElementTree as ET

from spreads.plugin import HookPlugin
from spreads.util import find_in_path, MissingDependencyException

if not find_in_path('scantailor-cli'):
    raise MissingDependencyException("Could not find executable"
                                     " `scantailor-cli` in $PATH. Please"
                                     " install the appropriate package(s)!")

logger = logging.getLogger('spreadsplug.scantailor')


class ScanTailorPlugin(HookPlugin):
    @classmethod
    def add_arguments(cls, command, parser):
        if command == "postprocess":
            parser.add_argument(
                "--auto", "-a", dest="autopilot", action="store_true",
                default=False,
                help="Don't prompt user to edit ScanTailor configuration")
            parser.add_argument(
                "--page-detection", "-pd", dest="page_detection",
                action="store_true", default=False,
                help="Use page for layout instead of content")

    def _generate_configuration(self, projectfile, img_dir, out_dir):
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
        logger.info("Generating ScanTailor configuration")
        filterconf = [self.config['scantailor'][x].get(bool)
                      for x in ('rotate', 'split_pages', 'deskew', 'content',
                                'auto_margins')]
        start_filter = filterconf.index(True)+1
        end_filter = len(filterconf) - list(reversed(filterconf)).index(True)+1
        marginconf = self.config['scantailor']['margins'].as_str_seq()
        generation_cmd = ['scantailor-cli',
                          '--start-filter={0}'.format(start_filter),
                          '--end-filter={0}'.format(end_filter),
                          '--layout=1.5',
                          '--dpi={0}'.format(self.config['device']['dpi']
                                             .get(int)),
                          '-o={0}'.format(projectfile)]
        page_detection = (
            self.config['scantailor']['detection'] == 'page'
            or ('page_detection' in self.config.keys()
                and self.config['page_detection'].get(bool))
        )
        if page_detection:
            generation_cmd.extend([
                '--enable-page-detection',
                '--disable-content-detection',
                '--enable-fine-tuning'
            ])
        else:
            generation_cmd.extend([
                '--margins-top={0}'.format(marginconf[0]),
                '--margins-right={0}'.format(marginconf[1]),
                '--margins-bottom={0}'.format(marginconf[2]),
                '--margins-left={0}'.format(marginconf[3]),
            ])
        generation_cmd.extend([img_dir, out_dir])
        logger.debug(" ".join(generation_cmd))
        subprocess.call(generation_cmd)

    def _split_configuration(self, projectfile, temp_dir):
        num_pieces = multiprocessing.cpu_count()
        tree = ET(file=projectfile)
        num_files = len(tree.findall('./files/file'))
        splitfiles = []
        files_per_job = int(math.ceil(float(num_files)/num_pieces))
        for idx in xrange(num_pieces):
            tree = ET(file=projectfile)
            root = tree.getroot()
            start = idx*files_per_job
            end = start + files_per_job
            if end > num_files:
                end = None
            for elem in ('files', 'images', 'pages',
                         'file-name-disambiguation'):
                elem_root = root.find(elem)
                to_keep = elem_root.getchildren()[start:end]
                to_remove = [x for x in elem_root.getchildren()
                             if not x in to_keep]
                for node in to_remove:
                    elem_root.remove(node)
            out_file = os.path.join(temp_dir,
                                    "{0}-{1}.ScanTailor".format(
                                    os.path.splitext(os.path.basename(
                                    projectfile))[0], idx))
            tree.write(out_file)
            splitfiles.append(out_file)
        return splitfiles

    def _generate_output(self, projectfile, out_dir):
        logger.debug("Generating output...")
        temp_dir = tempfile.mkdtemp(prefix="spreads.")
        split_config = self._split_configuration(projectfile, temp_dir)
        logger.debug("Launching those subprocesses!")
        processes = [subprocess.Popen(['scantailor-cli', '--start-filter=6',
                                       x, out_dir])
                     for x in split_config]
        while processes:
            for p in processes[:]:
                if p.poll() is not None:
                    processes.remove(p)
        shutil.rmtree(temp_dir)

    def process(self, path):
        autopilot = (self.config['scantailor']['autopilot']
                     .get(bool) or self.config['autopilot'].get(bool))
        if not autopilot and not find_in_path('scantailor'):
            raise MissingDependencyException(
                "Could not find executable `scantailor` in"
                " $PATH. Please install the appropriate"
                " package(s)!")
        projectfile = os.path.join(path, "{0}.ScanTailor".format(
            os.path.basename(path)))
        img_dir = os.path.join(path, 'raw')
        out_dir = os.path.join(path, 'done')

        if not os.path.exists(projectfile):
            self._generate_configuration(projectfile, img_dir, out_dir)
        if not autopilot:
            logger.info("Opening ScanTailor GUI for manual adjustment")
            subprocess.call(['scantailor', projectfile])
        logger.info("Generating output images from ScanTailor configuration.")
        self._generate_output(projectfile, out_dir)



from mock import MagicMock as Mock

import spreads
import spreadsplug.scantailor as scantailor

class TestScanTailor(object):
    def setUp(self):
        spreads.config.clear()
        spreads.config.read(user=False)
        self.stplug = scantailor.ScanTailorPlugin(spreads.config)

    def test_generate_configuration(self):
        scantailor.subprocess.call = Mock()
        # TODO: Setup up some config variables
        self.stplug._generate_configuration('/tmp/foo.st', '/tmp/raw',
                                            '/tmp/out')
        # TODO: Check the sp.call for the correct parameters

    def test_split_configuration(self):
        # TODO: Provide a sample configuration
        # TODO: Check for number of files
        # TODO: Check if the files have all the pages
        pass

    def test_generate_output(self):
        scantailor.subprocess.Popen = Mock()
        self.stplug._split_configuration = Mock()
        self.stplug._generate_output('/tmp/foo.st', '/tmp')


    def test_process(self):
        scantailor.subprocess.call = Mock()
        self.stplug._generate_configuration = Mock()
        self.stplug._generate_output = Mock()
        spreads.config['autopilot'] = True
        self.stplug.process('/tmp')
        assert scantailor.subprocess.call.call_count == 0
        spreads.config['autopilot'] = False
        self.stplug.process('/tmp')
        scantailor.subprocess.call.assert_called_with(
            ['scantailor', '/tmp/tmp.ScanTailor'])



#!/usr/bin/env python2.7
from setuptools import setup

setup(
    name='spreads',
    version='0.3.3',
    author='Johannes Baiter',
    author_email='johannes.baiter@gmail.com',
    packages=['spreads', 'spreadsplug', 'spreadsplug.dev'],
    include_package_data=True,
    scripts=['spread', ],
    url='http://github.com/jbaiter/spreads',
    license='MIT',
    description='Book digitization workflow assistant',
    long_description=open('README.rst').read(),
    install_requires=[
        "colorama >= 0.2.5",
        "pyusb >= 1.0.0a3",
        "pyptpchdk >= 0.2.1",
        "PyYAML >= 3.10",
        "Wand >= 0.3.1",
        "stevedore >= 0.9.1",
        "futures >= 2.1.4",
        "pexif >= 0.13",
        "Pillow >= 2.0.0",
    ],
    entry_points={
        'spreadsplug.devices': [
            'chdkcamera = spreadsplug.dev.chdkcamera:CHDKCameraDevice',
            'a2200      = spreadsplug.dev.chdkcamera:CanonA2200CameraDevice',
        ],
        'spreadsplug.hooks': [
            'combine    = spreadsplug.combine:CombinePlugin',
            'autorotate = spreadsplug.autorotate:AutoRotatePlugin',
            'scantailor = spreadsplug.scantailor:ScanTailorPlugin',
            'pdfbeads   = spreadsplug.pdfbeads:PDFBeadsPlugin',
            'djvubind   = spreadsplug.djvubind:DjvuBindPlugin',
            'colorcorrect = spreadsplug.colorcorrect:ColorCorrectionPlugin',
            'tesseract = spreadsplug.tesseract:TesseractPlugin',
            'gui        = spreadsplug.gui:GuiCommand',
        ],
    },
    test_suite='nose.collector',
)
